# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
from random import choices
from string import ascii_lowercase as ASCII_LOWERCASE


import numpy as np

from wulfric.cell._basic_manipulation import get_reciprocal
from wulfric.cell._voronoi import get_wigner_seitz_cell, get_lattice_points
from wulfric.constants import ATOM_COLORS
from wulfric.crystal._atoms import get_atom_species

try:
    import plotly.graph_objects as go  # noqa: F401
    from plotly.subplots import make_subplots

    PLOTLY_AVAILABLE = True
except ImportError:
    PLOTLY_AVAILABLE = False

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


_LEGEND_SETTINGS = {
    "top": dict(yanchor="bottom", y=1.0, xanchor="center", x=0.5),
    "bottom": dict(yanchor="top", y=0.0, xanchor="center", x=0.5),
    "left": dict(yanchor="middle", y=0.5, xanchor="right", x=0.0),
    "right": dict(yanchor="middle", y=0.5, xanchor="left", x=1.0),
}


# Source: https://gamedev.stackexchange.com/questions/38536/given-a-rgb-color-x-how-to-find-the-most-contrasting-color-y
def _get_good_contrast(hex_color):
    hex_color = hex_color[1:]
    R, G, B = [int(hex_color[i : i + 2], 16) / 255 for i in (0, 2, 4)]

    gamma = 2.2
    L = 0.2126 * R**gamma + 0.7152 * G**gamma + 0.0722 * B**gamma

    if L > 0.5**gamma:
        return "#000000"

    return "#FFFFFF"


class PlotlyEngine:
    r"""
    Plotting engine based on |plotly|_.

    Parameters
    ----------
    fig : plotly graph object, optional
        Figure to plot on. If not provided, then a new one is created as
        ``fig = go.Figure()``.
    _sphinx_gallery_fix : bool, default  False
        Fixes display issues when building documentation using sphinx gallery.
        Please, always ignore this argument

    Attributes
    ----------
    fig : plotly graph object
        Figure to plot on.

    Notes
    -----
    This class is a part of ``wulfric[visual]``
    """

    def __init__(self, fig=None, _sphinx_gallery_fix=False, rows=1, cols=1):
        if not PLOTLY_AVAILABLE:
            raise ImportError(
                'Plotly is not available. Please install it with "pip install plotly"'
            )

        if fig is None:
            fig = make_subplots(
                rows=rows,
                cols=cols,
                specs=[[{"type": "scene"} for _ in range(cols)] for _ in range(rows)],
            )
        self.fig = fig

        self.fig.update_layout(template="none")

        self._sphinx_gallery_fix = _sphinx_gallery_fix

        # Fix for plotly #7143
        self.x_range = {}
        self.y_range = {}
        self.z_range = {}

    # Fix for plotly #7143
    def _update_range(self, x_min, x_max, y_min, y_max, z_min, z_max, row=1, col=1):
        if (row, col) not in self.x_range:
            self.x_range[(row, col)] = np.array([x_min, x_max], dtype=float)
            self.y_range[(row, col)] = np.array([y_min, y_max], dtype=float)
            self.z_range[(row, col)] = np.array([z_min, z_max], dtype=float)
        else:
            self.x_range[(row, col)][0] = min(self.x_range[(row, col)][0], x_min)
            self.x_range[(row, col)][1] = max(self.x_range[(row, col)][1], x_max)
            self.y_range[(row, col)][0] = min(self.y_range[(row, col)][0], y_min)
            self.y_range[(row, col)][1] = max(self.y_range[(row, col)][1], y_max)
            self.z_range[(row, col)][0] = min(self.z_range[(row, col)][0], z_min)
            self.z_range[(row, col)][1] = max(self.z_range[(row, col)][1], z_max)

    # Fix for plotly #7143
    def _update_fig_aspect_range(self):
        for row, col in self.x_range:
            xlim = 1.05 * self.x_range[(row, col)]
            ylim = 1.05 * self.y_range[(row, col)]
            zlim = 1.05 * self.z_range[(row, col)]

            self.fig.update_scenes(
                xaxis=dict(range=xlim),
                yaxis=dict(range=ylim),
                zaxis=dict(range=zlim),
                aspectmode="manual",
                aspectratio=dict(
                    x=abs((xlim[1] - xlim[0]) / (zlim[1] - zlim[0])),
                    y=abs((ylim[1] - ylim[0]) / (zlim[1] - zlim[0])),
                    z=1,
                ),
                row=row,
                col=col,
            )

    def show(self, axes_visible=True, legend_position="top", **kwargs):
        r"""
        Shows the figure in the interactive mode.

        Parameters
        ----------
        axes_visible : bool, default True
            Whether to show axes.
        legend_position : str, default "top"
            Positions of the legend, case insensitive.
        **kwargs
            Passed directly to the |plotly-update-layout|_.
        """

        if not axes_visible:
            self.fig.update_scenes(
                xaxis_visible=False, yaxis_visible=False, zaxis_visible=False
            )

        legend_position = legend_position.lower()

        if legend_position not in list(_LEGEND_SETTINGS):
            raise ValueError(
                f"Supported values for legend_position are {list(_LEGEND_SETTINGS)}, got {legend_position}."
            )
        legend = _LEGEND_SETTINGS[legend_position]

        self.fig.update_layout(**kwargs, legend=legend)

        # Fix for plotly #7143
        self._update_fig_aspect_range()

        if self._sphinx_gallery_fix:
            return self.fig

        self.fig.show()

    def save(
        self,
        output_name="wulfric-plot.html",
        axes_visible=True,
        legend_position="top",
        kwargs_write_html=None,
        **kwargs,
    ):
        r"""
        Saves the figure in the html file.

        Parameters
        ----------
        output_name : str, default "lattice_graph.html"
            Name of the file to be saved. With extension.
        axes_visible : bool, default True
            Whether to show axes.
        legend_position : str, default "top"
            Positions of the legend, case insensitive.
        kwargs_write_html : dict, optional
            Passed directly to the |plotly-write-html|_.
        **kwargs
            Passed directly to the |plotly-update-layout|_.
        """

        if kwargs_write_html is None:
            kwargs_write_html = {}

        if not axes_visible:
            self.fig.update_scenes(
                xaxis_visible=False, yaxis_visible=False, zaxis_visible=False
            )

        legend_position = legend_position.lower()

        if legend_position not in list(_LEGEND_SETTINGS):
            raise ValueError(
                f"Supported values for legend_position are {list(_LEGEND_SETTINGS)}, got {legend_position}."
            )
        legend = _LEGEND_SETTINGS[legend_position]

        self.fig.update_layout(**kwargs, legend=legend)

        # Fix for plotly #7143
        self._update_fig_aspect_range()

        self.fig.write_html(output_name, **kwargs_write_html)

    def plot_points(
        self,
        points,
        colors="#000000",
        legend_label=None,
        legend_group=None,
        scale=1,
        row=1,
        col=1,
    ):
        r"""
        Plots a set of points.

        Parameters
        ----------
        points : (N, 3) |array-like|_
            Coordinates of the points.
        colors : str or list of str, default "#000000"
            Color of the line. Any value that is supported by |plotly|_.
        legend_label : str, optional
            Label of the line that is displayed in the figure.
        legend_group : str, optional
            Legend's group. If ``None``, then defaults to the random string of 10
            characters.
        scale : float, default 1
            Scale for the size of point's markers. Use ``scale>1`` to increase the size.
        row : int, default 1
            Row of the subplot.
        col : int, default 1
            Column of the subplot.
        """

        if legend_group is None:
            legend_group = "".join(choices(ASCII_LOWERCASE, k=10))

        points = np.array(points).T

        self.fig.add_traces(
            data=dict(
                type="scatter3d",
                mode="markers",
                legendgroup=legend_group,
                name=legend_label,
                showlegend=legend_label is not None,
                x=points[0],
                y=points[1],
                z=points[2],
                marker=dict(size=2 * scale, color=colors),
                hoverinfo="none",
            ),
            rows=row,
            cols=col,
        )

        # Fix for plotly #7143
        self._update_range(
            x_min=points[0].min(),
            x_max=points[0].max(),
            y_min=points[1].min(),
            y_max=points[1].max(),
            z_min=points[2].min(),
            z_max=points[2].max(),
            row=row,
            col=col,
        )

    def plot_line(
        self,
        start_point,
        end_point,
        color="#000000",
        legend_label=None,
        legend_group=None,
        row=1,
        col=1,
    ):
        r"""
        Plots a single line between ``start_point`` to ``end_point``.

        Parameters
        ----------
        start_point : (3, ) |array-like|_
            First end of the line.
        end_point : (3, ) |array-like|_
            Second point of the line.
        color : str, default "#000000"
            Color of the line. Any value that is supported by |plotly|_.
        legend_label : str, optional
            Label of the line that is displayed in the figure.
        legend_group : str, optional
            Legend's group. If ``None``, then defaults to the random string of 10
            characters.
        row : int, default 1
            Row of the subplot.
        col : int, default 1
            Column of the subplot.
        """

        if legend_group is None:
            legend_group = "".join(choices(ASCII_LOWERCASE, k=10))

        x, y, z = np.array([start_point, end_point]).T

        self.fig.add_traces(
            data=dict(
                type="scatter3d",
                mode="lines",
                legendgroup=legend_group,
                name=legend_label,
                showlegend=legend_label is not None,
                x=x,
                y=y,
                z=z,
                line=dict(color=color),
                hoverinfo="none",
            ),
            rows=row,
            cols=col,
        )

        # Fix for plotly #7143
        self._update_range(
            x_min=x.min(),
            x_max=x.max(),
            y_min=y.min(),
            y_max=y.max(),
            z_min=z.min(),
            z_max=z.max(),
            row=row,
            col=col,
        )

    def plot_vector(
        self,
        start_point,
        end_point,
        color="#000000",
        vector_label=None,
        legend_label=None,
        legend_group=None,
        row=1,
        col=1,
    ):
        r"""
        Plots a single vector pointing from ``start_point`` to ``end_point``.

        Parameters
        ----------
        start_point : (3, ) |array-like|_
            Start point of the vector.
        end_point : (3, ) |array-like|_
            End point of the vector.
        color : str, default "#000000"
            Color of the vector and its label. Any value that is supported by |plotly|_.
        vector_label : str, optional
            Label of the vector that is displayed in the figure.
        legend_label : str, optional
            Label for the legend. Entry in legend only showed if
            ``legend_label is not None``.
        legend_group : str, optional
            Legend's group. If ``None``, then defaults to the random string of 10
            characters.
        row : int, default 1
            Row of the subplot.
        col : int, default 1
            Column of the subplot.
        """

        if legend_group is None:
            legend_group = "".join(choices(ASCII_LOWERCASE, k=10))

        x, y, z = np.array([start_point, end_point]).T

        self.fig.add_traces(
            data=[
                dict(
                    type="scatter3d",
                    mode="lines",
                    legendgroup=legend_group,
                    name=legend_label,
                    showlegend=legend_label is not None,
                    x=x,
                    y=y,
                    z=z,
                    line=dict(color=color, width=3),
                    hoverinfo="none",
                ),
                dict(
                    type="cone",
                    legendgroup=legend_group,
                    showlegend=False,
                    x=[x[1]],
                    y=[y[1]],
                    z=[z[1]],
                    u=[0.4 * (x[1] - x[0])],
                    v=[0.4 * (y[1] - y[0])],
                    w=[0.4 * (z[1] - z[0])],
                    anchor="tip",
                    colorscale=[[0, color], [1, color]],
                    showscale=False,
                    hoverinfo="none",
                ),
            ],
            rows=row,
            cols=col,
        )
        if vector_label is not None:
            self.fig.add_traces(
                data=dict(
                    type="scatter3d",
                    mode="text",
                    legendgroup=legend_group,
                    showlegend=False,
                    x=[1.2 * (x[1] - x[0]) + x[0]],
                    y=[1.2 * (y[1] - y[0]) + y[0]],
                    z=[1.2 * (z[1] - z[0]) + z[0]],
                    marker=dict(size=0),
                    text=vector_label,
                    textfont=dict(size=12, color=color),
                    textposition="top center",
                    hoverinfo="none",
                ),
                rows=row,
                cols=col,
            )

            # Fix for plotly #7143
            self._update_range(
                x_min=1.2 * (x[1] - x[0]) + x[0],
                x_max=1.2 * (x[1] - x[0]) + x[0],
                y_min=1.2 * (y[1] - y[0]) + y[0],
                y_max=1.2 * (y[1] - y[0]) + y[0],
                z_min=1.2 * (z[1] - z[0]) + z[0],
                z_max=1.2 * (z[1] - z[0]) + z[0],
                row=row,
                col=col,
            )

        # Fix for plotly #7143
        self._update_range(
            x_min=x.min(),
            x_max=x.max(),
            y_min=y.min(),
            y_max=y.max(),
            z_min=z.min(),
            z_max=z.max(),
            row=row,
            col=col,
        )

    def plot_cell(
        self,
        cell,
        color="#000000",
        plot_vectors=True,
        vector_label="a",
        shift=(0, 0, 0),
        legend_label=None,
        legend_group=None,
        row=1,
        col=1,
    ):
        r"""
        Plots given ``cell`` as is.

        Parameters
        ----------
        cell : (3, 3) |array-like|_
            Matrix of a cell, rows are interpreted as vectors.
        color : str, default "#000000"
            Colour for the cell and the labels. Any value that is supported by |plotly|_.
        plot_vectors : bool, default True
            Whether to plot lattice vectors.
        vector_label : str, default "a"
            Vector's label, ignored if ``plot_vectors = False``.
        shift : (3, ) |array-like|_, default (0.0, 0.0, 0.0)
            Absolute coordinates of the corner of the cell, from which the three lattice
            vectors are plotted.
        legend_label : str, optional
            Label for the legend. Entry in legend only showed if
            ``legend_label is not None``.
        legend_group : str, optional
            Legend's group. If ``None``, then defaults to the random string of 10
            characters.
        row : int, default 1
            Row of the subplot.
        col : int, default 1
            Column of the subplot.
        """

        cell = np.array(cell)

        if legend_group is None:
            legend_group = "".join(choices(ASCII_LOWERCASE, k=10))

        shift = np.array(shift)

        # Plot vectors
        if plot_vectors:
            for i in range(3):
                self.plot_vector(
                    start_point=shift,
                    end_point=shift + cell[i],
                    color=color,
                    vector_label=f"{vector_label}{i + 1}",
                    legend_group=legend_group,
                    row=row,
                    col=col,
                )

        # Plot the cell borders
        for i in range(0, 3):
            j = (i + 1) % 3
            k = (i + 2) % 3
            self.plot_line(
                start_point=shift,
                end_point=shift + cell[i],
                color=color,
                legend_label=legend_label,
                legend_group=legend_group,
                row=row,
                col=col,
            )
            if legend_label is not None:
                legend_label = None
            self.plot_line(
                start_point=shift + cell[i],
                end_point=shift + cell[i] + cell[j],
                color=color,
                legend_group=legend_group,
                row=row,
                col=col,
            )
            self.plot_line(
                start_point=shift + cell[i],
                end_point=shift + cell[i] + cell[k],
                color=color,
                legend_group=legend_group,
                row=row,
                col=col,
            )
            self.plot_line(
                start_point=shift + cell[i] + cell[j],
                end_point=shift + cell[i] + cell[j] + cell[k],
                color=color,
                legend_group=legend_group,
                row=row,
                col=col,
            )

    def plot_wigner_seitz_cell(
        self,
        cell,
        plot_vectors=True,
        vector_label="a",
        color="#000000",
        shift=(0.0, 0.0, 0.0),
        legend_label=None,
        legend_group=None,
        row=1,
        col=1,
    ):
        r"""
        Plots Wigner-Seitz cell of the lattice spawned by the given ``cell``.

        Parameters
        ----------
        cell : (3, 3) |array-like|_
            Matrix of a cell, rows are interpreted as vectors.
        plot_vectors : bool, default True
            Whether to plot lattice vectors.
        vector_label : str, default "a"
            Vector's label, ignored if ``plot_vectors = False``.
        color : str, default "#000000"
            Colour for the cell and labels. Any value that is supported by |plotly|_.
        shift : (3, ) |array-like|_, default (0.0, 0.0, 0.0)
            Absolute coordinates of the center of the Wigner-Seitz cell.
        legend_label : str, optional
            Label for the legend. Entry in legend only showed if
            ``legend_label is not None``.
        legend_group : str, optional
            Legend's group. If ``None``, then defaults to the random string of 10
            characters.
        row : int, default 1
            Row of the subplot.
        col : int, default 1
            Column of the subplot.
        """

        cell = np.array(cell)

        if legend_group is None:
            legend_group = "".join(choices(ASCII_LOWERCASE, k=10))

        # Plot vectors
        if plot_vectors:
            for i in range(3):
                self.plot_vector(
                    start_point=shift,
                    end_point=shift + cell[i],
                    color=color,
                    vector_label=f"{vector_label}{i + 1}",
                    legend_group=legend_group,
                    row=row,
                    col=col,
                )

        vertices, edges = get_wigner_seitz_cell(cell=cell)
        for start_index, end_index in edges:
            self.plot_line(
                start_point=shift + vertices[start_index],
                end_point=shift + vertices[end_index],
                color=color,
                legend_label=legend_label,
                legend_group=legend_group,
                row=row,
                col=col,
            )
            if legend_label is not None:
                legend_label = None

    def plot_brillouin_zone(
        self,
        cell,
        plot_vectors=True,
        vector_label="b",
        color="#FF4D67",
        shift=(0.0, 0.0, 0.0),
        legend_label=None,
        legend_group=None,
        row=1,
        col=1,
    ):
        r"""

        Plots Brillouin zone.

        Parameters
        ----------
        cell : (3, 3) |array-like|_
            Matrix of a cell, rows are interpreted as vectors.
        plot_vectors : bool, default True
            Whether to plot lattice vectors.
        vector_label : str, default "b"
            Vector's label, ignored if ``plot_vectors = False``.
        color : str, default "#FF4D67"
            Colour for the Brillouin zone and labels. Any value that is supported by |plotly|_.
        shift : (3, ) |array-like|_, default (0.0, 0.0, 0.0)
            Absolute coordinates of the center of the Brillouin zone.
        legend_label : str, optional
            Label for the legend. Entry in legend only showed if
            ``legend_label is not None``.
        legend_group : str, optional
            Legend's group. If ``None``, then defaults to the random string of 10
            characters.
        row : int, default 1
            Row of the subplot.
        col : int, default 1
            Column of the subplot.
        """

        self.plot_wigner_seitz_cell(
            cell=get_reciprocal(cell),
            plot_vectors=plot_vectors,
            vector_label=vector_label,
            color=color,
            shift=shift,
            legend_label=legend_label,
            legend_group=legend_group,
            row=row,
            col=col,
        )

    def plot_kpath(
        self,
        kp,
        color="#000000",
        shift=(0.0, 0.0, 0.0),
        legend_label=None,
        legend_group=None,
        row=1,
        col=1,
    ):
        r"""
        Plots k-path in the reciprocal space.

        Parameters
        ----------
        kp : :py:class:`.Kpoints`
            K-points and k-path.
        color : str, default "#000000"
            Colour for the plot. Any value that is supported by |plotly|_.
        shift : (3, ) |array-like|_, default (0, 0, 0)
            Absolute coordinates of the shift in reciprocal space.
        legend_label : str, optional
            Label for the legend. Entry in legend only showed if
            ``legend_label is not None``.
        legend_group : str, optional
            Legend's group. If ``None``, then defaults to the random string of 10
            characters.
        row : int, default 1
            Row of the subplot.
        col : int, default 1
            Column of the subplot.
        """

        shift = np.array(shift)

        if legend_group is None:
            legend_group = "".join(choices(ASCII_LOWERCASE, k=10))

        for subpath in kp.path:
            xyz = []
            for i in range(len(subpath)):
                xyz.append(shift + kp.hs_coordinates[subpath[i]] @ kp.rcell)

            xyz = np.array(xyz).T
            self.fig.add_traces(
                data=dict(
                    type="scatter3d",
                    mode="lines",
                    name=legend_label,
                    legendgroup=legend_group,
                    showlegend=legend_label is not None,
                    x=xyz[0],
                    y=xyz[1],
                    z=xyz[2],
                    line=dict(width=3, color=color),
                    hoverinfo="none",
                ),
                rows=row,
                cols=col,
            )
            if legend_label is not None:
                legend_label = None

        # Fix for plotly #7143
        self._update_range(
            x_min=xyz[0].min(),
            x_max=xyz[0].max(),
            y_min=xyz[1].min(),
            y_max=xyz[1].max(),
            z_min=xyz[2].min(),
            z_max=xyz[2].max(),
            row=row,
            col=col,
        )

    def plot_kpoints(
        self,
        kp,
        color="#000000",
        shift=(0.0, 0.0, 0.0),
        legend_label=None,
        legend_group=None,
        scale=1,
        only_from_kpath=False,
        row=1,
        col=1,
    ):
        r"""
        Plots high-symmetry k-points in the reciprocal space.

        Parameters
        ----------
        kp : :py:class:`.Kpoints`
            K-points and k-path.
        color : str, default "#000000"
            Colour for the plot. Any value that is supported by |plotly|_.
        shift : (3, ) |array-like|_, default (0, 0, 0)
            Absolute coordinates of the shift in reciprocal space.
        legend_label : str, optional
            Label for the legend. Entry in legend only showed if
            ``legend_label is not None``.
        legend_group : str, optional
            Legend's group. If ``None``, then defaults to the random string of 10
            characters.
        scale : float, default 1
            Scale for the size of point's markers and text labels. Use ``scale>1`` to
            increase the size.
        only_from_kpath : bool, default False
            Whether to plot all pre-defined points or only the ones that included into the
            predefined k-path.
        row : int, default 1
            Row of the subplot.
        col : int, default 1
            Column of the subplot.
        """

        shift = np.array(shift)

        if legend_group is None:
            legend_group = "".join(choices(ASCII_LOWERCASE, k=10))

        p_abs = []
        p_rel = []
        labels = []
        path_points = []
        for subpath in kp.path:
            path_points += subpath
        for point in kp.hs_names:
            if not only_from_kpath or point in path_points:
                p_abs.append(shift + kp.hs_coordinates[point] @ kp.rcell)
                p_rel.append(kp.hs_coordinates[point])

                labels.append(point)

        p_abs = np.array(p_abs).T

        self.fig.add_traces(
            data=dict(
                type="scatter3d",
                mode="markers+text",
                legendgroup=legend_group,
                name=legend_label,
                showlegend=legend_label is not None,
                x=p_abs[0],
                y=p_abs[1],
                z=p_abs[2],
                marker=dict(size=5 * scale, color=color),
                text=labels,
                textposition="top center",
                textfont=dict(size=13 * scale, color=color),
                hovertext=p_rel,
                hoverinfo="text",
            ),
            rows=row,
            cols=col,
        )

        # Fix for plotly #7143
        self._update_range(
            x_min=p_abs[0].min(),
            x_max=p_abs[0].max(),
            y_min=p_abs[1].min(),
            y_max=p_abs[1].max(),
            z_min=p_abs[2].min(),
            z_max=p_abs[2].max(),
            row=row,
            col=col,
        )

    def plot_lattice(
        self,
        cell,
        color="#000000",
        range=(1, 1, 1),
        shift=(0, 0, 0),
        legend_label=None,
        legend_group=None,
        row=1,
        col=1,
    ):
        r"""
        Plots lattice points spawned by the given ``cell``.

        Parameters
        ----------
        cell : (3, 3) |array-like|_
            Matrix of a cell, rows are interpreted as vectors.
        color : str, default "#000000"
            Color of the points. Any value that is supported by |plotly|_.
        range : (3, ) tuple of int, default (1, 1, 1)
            How many lattice points to plot. All lattice points with relative coordinates
            ``r_1``, ``r_2``, ``r_3``, that fulfil

            * ``-range[0] <= r_1 <= range[0]``
            * ``-range[1] <= r_2 <= range[1]``
            * ``-range[2] <= r_3 <= range[2]``

            are plotted.
        shift : (3, ) |array-like|_, default (0, 0, 0)
            Absolute coordinates of the corner of the cell, from which the three lattice
            vectors are plotted.
        legend_label : str, optional
            Label for the legend. Entry in legend only showed if
            ``legend_label is not None``.
        legend_group : str, optional
            Legend's group. If ``None``, then defaults to the random string of 10
            characters.
        row : int, default 1
            Row of the subplot.
        col : int, default 1
            Column of the subplot.
        """

        cell = np.array(cell)

        if legend_group is None:
            legend_group = "".join(choices(ASCII_LOWERCASE, k=10))

        points = (
            get_lattice_points(cell=cell, range=range, relative=False, flat=True)
            + np.array(shift)[np.newaxis, :]
        )

        self.plot_points(
            points=points,
            colors=color,
            legend_label=legend_label,
            legend_group=legend_group,
            row=row,
            col=col,
        )

    def plot_atoms(
        self,
        cell,
        atoms,
        colors=None,
        legend_label=None,
        legend_group=None,
        shift=(0, 0, 0),
        scale=1,
        row=1,
        col=1,
    ):
        r"""
        Plots a set of atoms.

        Parameters
        ----------
        cell : (3, 3) |array-like|_
            Matrix of a cell, rows are interpreted as vectors.
        atoms : dict
            Dictionary with N atoms. Expected keys:

            *   "positions" : (N, 3) |array-like|_

                Positions of the atoms in the basis of lattice vectors (``cell``). In other
                words - relative coordinates of atoms.
            *   "names" : (N, ) list of str, optional
                See Notes

            *   "species" : (N, ) list of str, optional
                See Notes

        colors : str or list of str, optional
            Color of the atoms. Any value that is supported by |plotly|_. If ``None``,
            then color is deduced based on atoms's species.
        legend_label : str, optional
            Label of the line that is displayed in the figure.
        legend_group : str, optional
            Legend's group. If ``None``, then defaults to the random string of 10
            characters.
        scale : float, default 1
            Scale for the size of atoms's markers and text labels. Use ``scale>1`` to
            increase the size.
        row : int, default 1
            Row of the subplot.
        col : int, default 1
            Column of the subplot.

        Notes
        =====
        ``atoms["names"]`` is used to deduce atom's species if necessary via
        :py:func:`wulfric.crystal.get_atom_species`.

        ``atoms["species"] is used to define atom's colors if ``colors is None``
        """
        cell = np.array(cell)

        if legend_group is None:
            legend_group = "".join(choices(ASCII_LOWERCASE, k=10))

        points = atoms["positions"] @ cell + np.array(shift)[np.newaxis, :]

        points = points.T

        if "names" in atoms:
            names = atoms["names"]
        else:
            names = [f"X{i + 1}" for i in range(len(atoms["positions"]))]

        if colors is None:
            if "species" in atoms:
                species = atoms["species"]
            else:
                species = [
                    get_atom_species(name=name, raise_on_fail=False) for name in names
                ]

            colors = [ATOM_COLORS[_] for _ in species]

        text_color = [_get_good_contrast(color) for color in colors]

        self.fig.add_traces(
            data=dict(
                type="scatter3d",
                mode="markers+text",
                legendgroup=legend_group,
                name=legend_label,
                showlegend=legend_label is not None,
                x=points[0],
                y=points[1],
                z=points[2],
                text=names,
                marker=dict(size=14 * scale, color=colors),
                hoverinfo="none",
                textfont=dict(size=10 * scale, color=text_color),
                textposition="middle center",
            ),
            rows=row,
            cols=col,
        )

        # Fix for plotly #7143
        self._update_range(
            x_min=points[0].min(),
            x_max=points[0].max(),
            y_min=points[1].min(),
            y_max=points[1].max(),
            z_min=points[2].min(),
            z_max=points[2].max(),
            row=row,
            col=col,
        )


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
import numpy as np

from wulfric.cell._voronoi import _get_voronoi_cell


def test_voronoi_get_lattice_points_range():
    cell = np.array(
        [
            [2.8480, 0.0000, 0.0000],
            [0.0000, 2.8480, 0.0000],
            [1.4240, 1.4240, 1.4240],
        ]
    )
    vertices, edges = _get_voronoi_cell(cell)
    assert len(edges) == 36
    assert len(vertices) == 24



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================

import numpy as np

from wulfric._exceptions import ConventionNotSupported, PotentialBugError
from wulfric.crystal._crystal_validation import validate_atoms
from wulfric.cell._niggli import get_niggli
from wulfric.cell._basic_manipulation import get_reciprocal, get_params
from wulfric._spglib_interface import get_spglib_data, validate_spglib_data
from wulfric._syntactic_sugar import SyntacticSugar
from wulfric._numerical import compare_with_tolerance as cwt

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _hpkot_get_conventional_a(spglib_std_lattice):
    r"""
    Special case of hkpot convention and aP lattice

    Parameters
    ==========
    spglib_std_lattice : (3, 3) :numpy:`ndarray`
        Conventional cell found by spglib.

    Returns
    =======
    conv_cell : (3, 3) :numpy:`ndarray`
        Conventional cell as per the convention of HPKOT.
    """

    # Step 1 - get cell that is niggli-reduced in reciprocal space
    r_cell_step_1 = get_niggli(cell=get_reciprocal(spglib_std_lattice))
    cell_step_1 = get_reciprocal(r_cell_step_1)

    # Step 2
    dot_bc = abs(r_cell_step_1[1] @ r_cell_step_1[2])
    dot_ac = abs(r_cell_step_1[0] @ r_cell_step_1[2])
    dot_ab = abs(r_cell_step_1[0] @ r_cell_step_1[1])

    if dot_bc <= dot_ac and dot_bc <= dot_ab:
        matrix_to_2 = np.array(
            [
                [0, 0, 1],
                [1, 0, 0],
                [0, 1, 0],
            ]
        )
    elif dot_ac <= dot_bc and dot_ac <= dot_ab:
        matrix_to_2 = np.array(
            [
                [0, 1, 0],
                [0, 0, 1],
                [1, 0, 0],
            ]
        )
    elif dot_ab <= dot_ac and dot_ab <= dot_bc:
        matrix_to_2 = np.eye(3, dtype=float)
    else:
        raise PotentialBugError(
            '(convention="HPKOT"): aP lattice, step 2. Values of the dot products fall outside of three cases.'
        )

    cell_step_2 = matrix_to_2.T @ cell_step_1

    # Step 3
    _, _, _, r_alpha, r_beta, r_gamma = get_params(
        cell=get_reciprocal(cell=cell_step_2)
    )

    if (r_alpha < 90.0 and r_beta < 90.0 and r_gamma < 90.0) or (
        r_alpha >= 90.0 and r_beta >= 90.0 and r_gamma >= 90.0
    ):
        matrix_to_3 = np.eye(3, dtype=float)
    elif (r_alpha > 90.0 and r_beta > 90.0 and r_gamma < 90.0) or (
        r_alpha <= 90.0 and r_beta <= 90.0 and r_gamma >= 90.0
    ):
        matrix_to_3 = np.array(
            [
                [-1, 0, 0],
                [0, -1, 0],
                [0, 0, 1],
            ],
            dtype=float,
        )
    elif (r_alpha > 90.0 and r_beta < 90.0 and r_gamma > 90.0) or (
        r_alpha <= 90.0 and r_beta >= 90.0 and r_gamma <= 90.0
    ):
        matrix_to_3 = np.array(
            [
                [-1, 0, 0],
                [0, 1, 0],
                [0, 0, -1],
            ],
            dtype=float,
        )
    elif (r_alpha < 90.0 and r_beta > 90.0 and r_gamma > 90.0) or (
        r_alpha >= 90.0 and r_beta <= 90.0 and r_gamma <= 90.0
    ):
        matrix_to_3 = np.array(
            [
                [1, 0, 0],
                [0, -1, 0],
                [0, 0, -1],
            ],
            dtype=float,
        )
    else:
        raise PotentialBugError(
            '(convention="HPKOT"): aP lattice, step 3. Values of the reciprocal angles fall outside of four cases.'
        )

    return matrix_to_3.T @ cell_step_2


def _sc_get_conventional_oPFI(spglib_std_lattice):
    r"""
    Case of SC convention and oP, oF or oI lattice.

    Choose the cell with the lattice parameters that satisfy

    * ``a < b < c``
    * ``alpha = beta = gamma = 90``

    Parameters
    ==========
    spglib_std_lattice : (3, 3) :numpy:`ndarray`
        Conventional cell found by spglib.

    Returns
    =======
    conv_cell : (3, 3) :numpy:`ndarray`
        Conventional cell as per the convention of SC.
    """

    a, b, c, _, _, _ = get_params(cell=spglib_std_lattice)

    if a < b < c:  # No change
        matrix = np.eye(3, dtype=float)
    elif a < c < b:  # -> -a1, -a3, -a2
        matrix = np.array(
            [
                [-1, 0, 0],
                [0, 0, -1],
                [0, -1, 0],
            ],
            dtype=float,
        )
    elif b < a < c:  # -> -a2, -a1, -a3
        matrix = np.array(
            [
                [0, -1, 0],
                [-1, 0, 0],
                [0, 0, -1],
            ],
            dtype=float,
        )
    elif b < c < a:  # -> a2, a3, a1
        matrix = np.array(
            [
                [0, 0, 1],
                [1, 0, 0],
                [0, 1, 0],
            ],
            dtype=float,
        )
    elif c < a < b:  # -> a3, a1, a2
        matrix = np.array(
            [
                [0, 1, 0],
                [0, 0, 1],
                [1, 0, 0],
            ],
            dtype=float,
        )
    elif c < b < a:  # -> -a3, -a2, -a1
        matrix = np.array(
            [
                [0, 0, -1],
                [0, -1, 0],
                [-1, 0, 0],
            ],
            dtype=float,
        )
    else:
        raise PotentialBugError(
            '(convention="SC"): oP, oF, oI lattices. Length of the lattice vectors fall outside of six cases.'
        )

    return matrix.T @ spglib_std_lattice


def _sc_get_conventional_oC(spglib_std_lattice):
    r"""
    Case of SC convention and oC lattice.

    Choose the cell with the lattice parameters that satisfy

    * ``a < b``
    * ``alpha = beta = gamma = 90``

    Parameters
    ==========
    spglib_std_lattice : (3, 3) :numpy:`ndarray`
        Conventional cell found by spglib.

    Returns
    =======
    conv_cell : (3, 3) :numpy:`ndarray`
        Conventional cell as per the convention of SC.
    """

    a, b, _, _, _, _ = get_params(cell=spglib_std_lattice)

    if a < b:  # No change
        matrix = np.eye(3, dtype=float)
    # Have to keep a3 in place
    elif b < a:  # -> a2, -a1, a3
        matrix = np.array(
            [
                [0, -1, 0],
                [1, 0, 0],
                [0, 0, 1],
            ],
            dtype=float,
        )
    else:
        raise PotentialBugError(
            '(convention="SC"): oC lattices. Length of the lattice vectors fall outside of two cases.'
        )

    return matrix.T @ spglib_std_lattice


def _sc_get_conventional_m(spglib_std_lattice, centring_type, angle_tolerance=1e-4):
    r"""
    Case of SC convention and m lattice.

    Choose the cell with the lattice parameters that satisfy

    * ``b <= c``
    * ``alpha < 90``

    Parameters
    ==========
    spglib_std_lattice : (3, 3) :numpy:`ndarray`
        Conventional cell found by spglib.
    centring_type : str
        CEntring type of the lattice either "P" or "C".
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance parameter for comparing two angles, given in degrees.

    Returns
    =======
    conv_cell : (3, 3) :numpy:`ndarray`
        Conventional cell as per the convention of SC.
    """

    # Step 1, make sure that alpha is not a 90 angle
    _, _, _, alpha, beta, gamma = get_params(cell=spglib_std_lattice)
    if cwt(beta, "==", 90.0, eps=angle_tolerance) and cwt(
        gamma, "==", 90.0, eps=angle_tolerance
    ):  # No change
        matrix_to_1 = np.eye(3, dtype=float)
    elif cwt(alpha, "==", 90.0, eps=angle_tolerance) and cwt(
        beta, "==", 90.0, eps=angle_tolerance
    ):  # -> a3, a1, a2
        matrix_to_1 = np.array(
            [
                [0, 1, 0],
                [0, 0, 1],
                [1, 0, 0],
            ]
        )
        if centring_type == "C":
            raise PotentialBugError(
                '(convention="SC"), MCLC lattice, step 1, case 2. This case should never be tried for MCLC lattice.'
            )
    # Have to keep a3 in place for MCLC lattices
    elif cwt(alpha, "==", 90.0, eps=angle_tolerance) and cwt(
        gamma, "==", 90.0, eps=angle_tolerance
    ):  # -> a2, -a1, a3
        matrix_to_1 = np.array(
            [
                [0, -1, 0],
                [1, 0, 0],
                [0, 0, 1],
            ]
        )
    else:
        raise PotentialBugError(
            '(convention="SC"): MCL or MCLC lattice, step 1. Angles fall outside of the three cases.'
        )
    cell_step_1 = matrix_to_1.T @ spglib_std_lattice

    # Step 2, make sure that b <= c (only for MCL, not for MCLC)
    _, b, c, _, _, _ = get_params(cell=cell_step_1)
    if b <= c or centring_type == "C":
        matrix_to_2 = np.eye(3, dtype=float)
    else:  # -> -a1, a3, a2
        matrix_to_2 = np.array(
            [
                [-1, 0, 0],
                [0, 0, 1],
                [0, 1, 0],
            ],
            dtype=float,
        )
    cell_step_2 = matrix_to_2.T @ cell_step_1

    # Step 3, make sure that alpha < 90
    _, _, _, alpha, _, _ = get_params(cell=cell_step_2)
    if cwt(alpha, "<=", 90.0, eps=angle_tolerance):  # No change
        matrix_to_3 = np.eye(3, dtype=float)
    # Have to keep a3 in place for MCLC lattices
    else:  # -> -a1, -a2, a3
        matrix_to_3 = np.array(
            [
                [-1, 0, 0],
                [0, -1, 0],
                [0, 0, 1],
            ]
        )

    return matrix_to_3.T @ cell_step_2


def _sc_get_conventional_a(spglib_std_lattice, angle_tolerance=1e-4):
    r"""
    Case of SC convention and a lattice.

    Choose the cell with the lattice parameters that satisfy

    * ``b <= c``
    * ``alpha < 90``

    Parameters
    ==========
    spglib_std_lattice : (3, 3) :numpy:`ndarray`
        Conventional cell found by spglib.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance parameter for comparing two angles, given in degrees.

    Returns
    =======
    conv_cell : (3, 3) :numpy:`ndarray`
        Conventional cell as per the convention of SC.
    """

    # Compute reciprocal cell
    r_cell = get_reciprocal(cell=spglib_std_lattice)

    # Step 1
    _, _, _, r_alpha, r_beta, r_gamma = get_params(cell=r_cell)

    if (r_alpha < 90.0 and r_beta < 90.0 and r_gamma < 90.0) or (
        r_alpha >= 90.0 and r_beta >= 90.0 and r_gamma >= 90.0
    ):
        matrix_to_1 = np.eye(3, dtype=float)
    elif (r_alpha > 90.0 and r_beta > 90.0 and r_gamma < 90.0) or (
        r_alpha <= 90.0 and r_beta <= 90.0 and r_gamma >= 90.0
    ):
        matrix_to_1 = np.array(
            [
                [-1, 0, 0],
                [0, -1, 0],
                [0, 0, 1],
            ],
            dtype=float,
        )
    elif (r_alpha > 90.0 and r_beta < 90.0 and r_gamma > 90.0) or (
        r_alpha <= 90.0 and r_beta >= 90.0 and r_gamma <= 90.0
    ):
        matrix_to_1 = np.array(
            [
                [-1, 0, 0],
                [0, 1, 0],
                [0, 0, -1],
            ],
            dtype=float,
        )
    elif (r_alpha < 90.0 and r_beta > 90.0 and r_gamma > 90.0) or (
        r_alpha >= 90.0 and r_beta <= 90.0 and r_gamma <= 90.0
    ):
        matrix_to_1 = np.array(
            [
                [1, 0, 0],
                [0, -1, 0],
                [0, 0, -1],
            ],
            dtype=float,
        )
    else:
        raise PotentialBugError(
            '(convention="SC"): aP lattice, step 1. Values of the reciprocal angles fall outside of four cases.'
        )
    r_cell_step_1 = matrix_to_1.T @ r_cell

    # Step 2
    # Note np.linalg.inv(matrix_to_1) == matrix_to_1.T
    _, _, _, r_alpha, r_beta, r_gamma = get_params(cell=r_cell_step_1)

    if (
        r_gamma == min(r_alpha, r_beta, r_gamma)
        and cwt(r_gamma, ">=", 90.0, eps=angle_tolerance)
        or (
            r_gamma == max(r_alpha, r_beta, r_gamma)
            and cwt(r_gamma, "<=", 90.0, eps=angle_tolerance)
        )
    ):
        matrix_to_2 = np.eye(3, dtype=float)
    elif (
        r_beta == min(r_alpha, r_beta, r_gamma)
        and cwt(r_beta, ">=", 90.0, eps=angle_tolerance)
        or (
            r_beta == max(r_alpha, r_beta, r_gamma)
            and cwt(r_beta, "<=", 90.0, eps=angle_tolerance)
        )
    ):
        matrix_to_2 = np.array(
            [
                [0, 1, 0],
                [0, 0, 1],
                [1, 0, 0],
            ],
            dtype=float,
        )
    elif (
        r_alpha == min(r_alpha, r_beta, r_gamma)
        and cwt(r_alpha, ">=", 90.0, eps=angle_tolerance)
        or (
            r_alpha == max(r_alpha, r_beta, r_gamma)
            and cwt(r_alpha, "<=", 90.0, eps=angle_tolerance)
        )
    ):
        matrix_to_2 = np.array(
            [
                [0, 0, 1],
                [1, 0, 0],
                [0, 1, 0],
            ],
            dtype=float,
        )
    else:
        raise PotentialBugError(
            '(convention="SC"): aP lattice, step 2. Values of the reciprocal angles fall outside of four cases.'
        )

    return matrix_to_2.T @ get_reciprocal(cell=r_cell_step_1)


def _sc_get_conventional_hR(
    spglib_primitive_cell, angle_tolerance=1e-4, distance_tolerance=1e-5
):
    r"""
    Computes conventional cell for the case of SC and hR lattice.

    It checks that

    * All three angles between the lattice vectors are equal

    * All lattice vectors have the same length

    Parameters
    ==========
    spglib_primitive_cell : (3, 3) :numpy:`ndarray`
        Primitive cell found by spglib.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance parameter for comparing two angles, given in degrees.
    distance_tolerance : float, default :math:`10^{-5}`
        Tolerance parameter for comparing two linear variables.

    Returns
    =======
    conv_cell : (3, 3) :numpy:`ndarray`
        Conventional cell as per the convention of SC.
    """

    a, b, c, alpha, beta, gamma = get_params(cell=spglib_primitive_cell)

    if (
        cwt(a, "!=", b, eps=distance_tolerance)
        or cwt(a, "!=", c, eps=distance_tolerance)
        or cwt(b, "!=", c, eps=distance_tolerance)
    ):
        raise PotentialBugError(
            f'(convention="SC"): hR lattice. Lattice vectors have different lengths with the precision of {distance_tolerance:.5e}'
        )

    if cwt(alpha, "==", beta, eps=angle_tolerance) and cwt(
        alpha, "==", gamma, eps=angle_tolerance
    ):
        matrix = np.eye(3, dtype=float)
    elif cwt(180 - alpha, "==", beta, eps=angle_tolerance) and cwt(
        beta, "==", gamma, eps=angle_tolerance
    ):  # -> a1, -a2, -a3
        matrix = np.array(
            [
                [1, 0, 0],
                [0, -1, 0],
                [0, 0, -1],
            ]
        )
    elif cwt(alpha, "==", 180 - beta, eps=angle_tolerance) and cwt(
        beta, "==", gamma, eps=angle_tolerance
    ):  # -> -a1, a2, -a3
        matrix = np.array(
            [
                [-1, 0, 0],
                [0, 1, 0],
                [0, 0, -1],
            ]
        )
    elif cwt(alpha, "==", beta, eps=angle_tolerance) and cwt(
        beta, "==", 180 - gamma, eps=angle_tolerance
    ):  # -> -a1, -a2, a3
        matrix = np.array(
            [
                [-1, 0, 0],
                [0, -1, 0],
                [0, 0, 1],
            ]
        )
    else:
        raise PotentialBugError(
            '(convention="SC"): hR lattice. Angles can not be made equal.'
        )

    return matrix.T @ spglib_primitive_cell


def _sc_get_conventional_no_hR(
    spglib_std_lattice,
    crystal_family,
    centring_type,
    angle_tolerance=1e-4,
):
    r"""
    Computes conventional lattice for the case of SC.

    Note: do not process hR, as it is treated separately.

    Parameters
    ==========
    spglib_std_lattice : (3, 3) :numpy:`ndarray`
        Conventional cell found by spglib.
    crystal_family : str
    centring_type : str
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance parameter for comparing two angles, given in degrees.

    Returns
    =======
    conv_cell : (3, 3) :numpy:`ndarray`
        Conventional cell as per the convention of SC.
    """
    # Run over possible crystal families but hR
    if crystal_family in ["c", "t"] or (crystal_family == "h" and centring_type == "P"):
        conv_cell = spglib_std_lattice
    elif crystal_family == "o":
        if centring_type in ["P", "F", "I"]:
            conv_cell = _sc_get_conventional_oPFI(spglib_std_lattice=spglib_std_lattice)
        elif centring_type == "C":
            conv_cell = _sc_get_conventional_oC(spglib_std_lattice=spglib_std_lattice)
        elif centring_type == "A":
            # Make it C-centered
            # a1, a2, a3 -> a2, a3, a1
            # and treat like one
            matrix = np.array(
                [
                    [0, 0, 1],
                    [1, 0, 0],
                    [0, 1, 0],
                ]
            )
            conv_cell = _sc_get_conventional_oC(
                spglib_std_lattice=matrix.T @ spglib_std_lattice
            )
        else:
            raise PotentialBugError(
                f'(convention="sc"): crystal family "o". Unexpected centring type "{centring_type}".'
            )
    elif crystal_family == "m":
        conv_cell = _sc_get_conventional_m(
            spglib_std_lattice=spglib_std_lattice,
            centring_type=centring_type,
            angle_tolerance=angle_tolerance,
        )
    elif crystal_family == "a":
        conv_cell = _sc_get_conventional_a(
            spglib_std_lattice=spglib_std_lattice, angle_tolerance=angle_tolerance
        )
    else:
        raise PotentialBugError(
            f'(convention="sc"): unexpected crystal family "{crystal_family}".'
        )

    return conv_cell


def get_conventional(cell, atoms, convention="HPKOT", spglib_data=None):
    r"""
    Return conventional cell and atoms associated with the given ``cell`` and ``atoms``.

    Parameters
    ==========
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.
    atoms : dict
        Dictionary with N atoms. Expected keys:

        *   "positions" : (N, 3) |array-like|_

            Positions of the atoms in the basis of lattice vectors (``cell``). In other
            words - relative coordinates of atoms.
        *   "names" : (N, ) list of str, optional

            See Notes
        *   "species" : (N, ) list of str, optional

            See Notes
        *   "spglib_types" : (N, ) list of int, optional

            See Notes

        .. hint::
            Pass ``atoms = dict(positions=[[0, 0, 0]], spglib_types=[1])`` if you would
            like to interpret the ``cell`` alone (effectively assuming that the ``cell``
            is a primitive one).

    convention : str, default "HPKOT"
        Convention for the definition of the conventional cell. Case-insensitive.
        Supported:

        * "HPKOT" for [1]_
        * "SC" for [2]_
        * "spglib" for |spglib|_ [3]_

    spglib_data : :py:class:`.SyntacticSugar`, optional
        If you need more control on the parameters passed to the spglib, then
        you can get ``spglib_data`` manually and pass it to this function.
        Use wulfric's interface to |spglib|_ as

        .. code-block:: python

            spglib_data = wulfric.get_spglib_data(...)

        using the same ``cell`` and ``atoms["positions"]`` that you are passing to this
        function.

    Returns
    =======
    conventional_cell : (3, 3) :numpy:`ndarray`
        Conventional cell.
    conventional_atoms : dict
        Dictionary of atoms of the conventional cell. Has all the same keys as the
        original ``atoms``. The values of each key are updated in such a way that
        ``conventional_cell`` with ``conventional_atoms`` describe the same crystal (and
        in the same spatial orientation) as ``cell`` with ``atoms``. It has all keys as
        in ``atoms``. Additional key ``"spglib_types"`` is added if it was not present in
        ``atoms``.

    See Also
    ========
    :ref:`user-guide_conventions_which-cell`
    wulfric.crystal.get_primitive
    wulfric.get_spglib_data


    Notes
    =====
    |spglib|_ uses ``types`` to distinguish the atoms. To see how wulfric deduces the
    ``types`` for given atoms see :py:func:`wulfric.get_spglib_types`.

    If two atoms ``i`` and ``j`` have the same spglib_type (i.e.
    ``atoms["spglib_types"][i] == atoms["spglib_types"][j]``), but they have different
    property that is stored in ``atoms[key]`` (i.e ``atoms[key][i] != atoms[key][j]``),
    then those two atoms are considered equal. In the returned ``conventional_atoms``
    the value of the ``conventional_atoms[key]`` are populated based on the *last* found
    atom in ``atoms`` with each for spglib_type. This rule do not apply to the "positions"
    key.


    References
    ==========
    .. [1] Hinuma, Y., Pizzi, G., Kumagai, Y., Oba, F. and Tanaka, I., 2017.
           Band structure diagram paths based on crystallography.
           Computational Materials Science, 128, pp.140-184.
    .. [2] Setyawan, W. and Curtarolo, S., 2010.
           High-throughput electronic band structure calculations: Challenges and tools.
           Computational materials science, 49(2), pp. 299-312.
    .. [3] Togo, A., Shinohara, K. and Tanaka, I., 2024.
           Spglib: a software library for crystal symmetry search.
           Science and Technology of Advanced Materials: Methods, 4(1), p.2384822.
    """

    # Validate that the atoms dictionary is what expected of it
    validate_atoms(atoms=atoms, required_keys=["positions"], raise_errors=True)

    # Call spglib
    if spglib_data is None:
        spglib_data = get_spglib_data(cell=cell, atoms=atoms)
    # Or check that spglib_data were *most likely* produced via wulfric's interface
    elif not isinstance(spglib_data, SyntacticSugar):
        raise TypeError(
            f"Are you sure that spglib_data were produced via wulfric's interface? Expected SyntacticSugar, got {type(spglib_data)}."
        )
    # Validate that user-provided spglib_data match user-provided structure
    else:
        validate_spglib_data(cell=cell, atoms=atoms, spglib_data=spglib_data)

    # Define conventional cell, positions and types
    convention = convention.lower()
    if convention == "spglib" or convention == "hpkot":
        if convention == "hpkot" and spglib_data.crystal_family == "a":
            # Find conventional cell and update atom positions
            conv_cell = _hpkot_get_conventional_a(
                spglib_std_lattice=spglib_data.conventional_cell
            )
            # Compute relative positions with respect to the new cell
            # relative spglib -> Cartesian -> relative HPKOT
            conv_positions = (
                spglib_data.conventional_positions
                @ spglib_data.conventional_cell
                @ np.linalg.inv(conv_cell)
            )
        else:
            conv_cell = spglib_data.conventional_cell
            conv_positions = spglib_data.conventional_positions

        conv_types = spglib_data.conventional_types

    elif convention == "sc":
        # Treat hR in a special way, as it changes the volume of the cell
        if spglib_data.crystal_family == "h" and spglib_data.centring_type == "R":
            # Fix potential convention mismatch
            conv_cell = _sc_get_conventional_hR(
                spglib_primitive_cell=spglib_data.primitive_cell,
                angle_tolerance=spglib_data.angle_tolerance
                if spglib_data.angle_tolerance != -1
                else 1e-4,
                distance_tolerance=spglib_data.symprec,
            )

            # Compute relative positions with respect to the new cell
            # relative spglib -> Cartesian -> relative SC
            conv_positions = (
                spglib_data.primitive_positions
                @ spglib_data.primitive_cell
                @ np.linalg.inv(conv_cell)
            )

            conv_types = spglib_data.primitive_types

        # In other cases the volume of the cell does not change
        else:
            conv_cell = _sc_get_conventional_no_hR(
                spglib_std_lattice=spglib_data.conventional_cell,
                crystal_family=spglib_data.crystal_family,
                centring_type=spglib_data.centring_type,
                angle_tolerance=spglib_data.angle_tolerance
                if spglib_data.angle_tolerance != -1
                else 1e-4,
            )

            # Compute relative positions with respect to the new cell
            # relative spglib -> Cartesian -> relative SC
            conv_positions = (
                spglib_data.conventional_positions
                @ spglib_data.conventional_cell
                @ np.linalg.inv(conv_cell)
            )

            conv_types = spglib_data.conventional_types

    else:
        raise ConventionNotSupported(
            convention, supported_conventions=["HPKOT", "SC", "spglib"]
        )

    # Create conventional atoms
    conv_atoms = dict(positions=conv_positions)

    # Get mapping from original atoms to conventional ones through types
    types_mapping = {
        type_index: index for index, type_index in enumerate(spglib_data.original_types)
    }

    # Populate conv_atoms with all keys that have been defined in the original atoms.
    for key in atoms:
        if key != "positions":
            conv_atoms[key] = []
            for type_index in conv_types:
                conv_atoms[key].append(atoms[key][types_mapping[type_index]])

    # Add spglib_types to new atoms if necessary
    if "spglib_types" not in conv_atoms:
        conv_atoms["spglib_types"] = conv_types

    return conv_cell, conv_atoms


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
from math import cos, sin

import numpy as np

from wulfric._numerical import compare_with_tolerance as cwt
from wulfric.cell._basic_manipulation import get_params, get_reciprocal
from wulfric.constants._numerical import TORADIANS
from wulfric.crystal._crystal_validation import validate_atoms
from wulfric._spglib_interface import get_spglib_data, validate_spglib_data
from wulfric._syntactic_sugar import SyntacticSugar
from wulfric.constants._sc_convention import SC_BRAVAIS_LATTICE_SHORT_NAMES
from wulfric.crystal._conventional import get_conventional
from wulfric.crystal._primitive import get_primitive

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _BCT_variation(conv_a: float, conv_c: float):
    r"""
    Two variations of the BCT lattice.

    Condition :math:`a \ne c` is assumed.

    :math:`\text{BCT}_1: c < a` and :math:`\text{BCT}_2: c > a`

    Parameters
    ----------
    conv_a : float
        Length of the :math:`a_1 == a_2` vector of the conventional cell.
    conv_c : float
        Length of the :math:`a_3` vector of the conventional cell.

    Returns
    -------
    variation : str
        Variation of the lattice. "BCT1" or "BCT2".

    Raises
    ------
    ValueError
        If :math:`a == c`.
    """
    if conv_a > conv_c:
        return "BCT1"
    elif conv_a < conv_c:
        return "BCT2"
    else:
        raise ValueError('(convention="SC"): BCT variation). a == c')


def _ORCF_variation(
    conv_a: float, conv_b: float, conv_c: float, distance_tolerance=1e-8
):
    r"""
    Three variations of the ORCF lattice.

    Ordering :math:`a < b < c` is assumed.

    :math:`\text{ORCF}_1: \dfrac{1}{a^2} > \dfrac{1}{b^2} + \dfrac{1}{c^2}`,
    :math:`\text{ORCF}_2: \dfrac{1}{a^2} < \dfrac{1}{b^2} + \dfrac{1}{c^2}`,
    :math:`\text{ORCF}_3: \dfrac{1}{a^2} = \dfrac{1}{b^2} + \dfrac{1}{c^2}`,

    Parameters
    ----------
    conv_a : float
        Length of the :math:`a_1` vector of the conventional cell.
    conv_b : float
        Length of the :math:`a_2` vector of the conventional cell.
    conv_c : float
        Length of the :math:`a_3` vector of the conventional cell.
    distance_tolerance : float, default :math:`10^{-5}`
        Tolerance parameter for comparing two linear variables.

    Returns
    -------
    variation : str
        Variation of the lattice. "ORCF1", "ORCF2" or "ORCF3".

    Raises
    ------
    ValueError
        If :math:`a < b < c` is not satisfied.
    """
    if cwt(conv_a, ">=", conv_b, eps=distance_tolerance) or cwt(
        conv_b, ">=", conv_c, eps=distance_tolerance
    ):
        raise ValueError(
            f'(convention="SC"): ORCF variation. a < b < c is not satisfied with {distance_tolerance} tolerance.'
        )

    expression = 1 / conv_a**2 - 1 / conv_b**2 - 1 / conv_c**2
    if cwt(expression, "==", 0, eps=distance_tolerance):
        return "ORCF3"
    elif cwt(expression, ">", 0, eps=distance_tolerance):
        return "ORCF1"
    elif cwt(expression, "<", 0, eps=distance_tolerance):
        return "ORCF2"


def _RHL_variation(conv_alpha: float, angle_tolerance=1e-4):
    r"""
    Two variations of the RHL lattice.

    Condition :math:`\alpha \ne 90^{\circ}` is assumed.

    :math:`\text{RHL}_1 \alpha < 90^{\circ}`,
    :math:`\text{RHL}_2 \alpha > 90^{\circ}`

    Parameters
    ----------
    conv_alpha : float
        Angle between vectors :math:`a_1` and :math:`a_2` of the conventional cell in
        degrees.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance parameter for comparing two angles, given in degrees.

    Returns
    -------
    variation : str
        Variation of the lattice. Either "RHL1" or "RHL2".

    Raises
    ------
    ValueError
        If :math:`\alpha == 90^{\circ}` with given tolerance ``eps``.
    """
    if cwt(conv_alpha, "<", 90, eps=angle_tolerance):
        return "RHL1"
    elif cwt(conv_alpha, ">", 90, eps=angle_tolerance):
        return "RHL2"
    else:
        raise ValueError(
            f'(convention="SC"): RHL variation. alpha == 90 with {angle_tolerance} tolerance.'
        )


def _MCLC_variation(
    conv_a: float,
    conv_b: float,
    conv_c: float,
    conv_alpha: float,
    prim_k_gamma: float,
    distance_tolerance=1e-8,
    angle_tolerance=1e-4,
):
    r"""
    Five variations of the MCLC lattice.

    :math:`\alpha < 90^{\circ}` is checked.

    :math:`\text{MCLC}_1: k_{\gamma} > 90^{\circ}`,
    :math:`\text{MCLC}_2: k_{\gamma} = 90^{\circ}`,
    :math:`\text{MCLC}_3: k_{\gamma} < 90^{\circ}, \dfrac{b\cos(\alpha)}{c} + \dfrac{b^2\sin(\alpha)^2}{a^2} < 1`
    :math:`\text{MCLC}_4: k_{\gamma} < 90^{\circ}, \dfrac{b\cos(\alpha)}{c} + \dfrac{b^2\sin(\alpha)^2}{a^2} = 1`
    :math:`\text{MCLC}_5: k_{\gamma} < 90^{\circ}, \dfrac{b\cos(\alpha)}{c} + \dfrac{b^2\sin(\alpha)^2}{a^2} > 1`

    Parameters
    ----------
    conv_a : float
        Length of the :math:`a_1` vector of the conventional cell.
    conv_b : float
        Length of the :math:`a_2` vector of the conventional cell.
    conv_c : float
        Length of the :math:`a_3` vector of the conventional cell.
    conv_alpha : float
        Angle between vectors :math:`a_2` and :math:`a_3` of the conventional cell in
        degrees.
    k_gamma : float
        Angle between reciprocal vectors :math:`b_1` and :math:`b_2`. In degrees.
    distance_tolerance : float, default :math:`10^{-5}`
        Tolerance parameter for comparing two linear variables.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance parameter for comparing two angles, given in degrees.

    Returns
    -------
    variation : str
        Variation of the lattice.
        Either "MCLC1", "MCLC2", "MCLC3", "MCLC4" or "MCLC5".

    Raises
    ------
    ValueError
        If :math:`\alpha > 90^{\circ}` with given tolerance ``angle_tolerance``.
    """

    if cwt(conv_alpha, ">", 90, eps=angle_tolerance):
        raise ValueError(
            f'(convention="SC"): MCLC variation. alpha > 90 with {angle_tolerance} tolerance:\n  alpha = {conv_alpha}\n'
        )

    conv_alpha *= TORADIANS

    if cwt(prim_k_gamma, "==", 90, eps=angle_tolerance):
        return "MCLC2"
    elif cwt(prim_k_gamma, ">", 90, eps=angle_tolerance):
        return "MCLC1"
    elif cwt(prim_k_gamma, "<", 90, eps=angle_tolerance):
        expression = (
            conv_b * cos(conv_alpha) / conv_c
            + conv_b**2 * sin(conv_alpha) ** 2 / conv_a**2
        )
        if cwt(expression, "==", 1, eps=distance_tolerance):
            return "MCLC4"
        elif cwt(expression, "<", 1, eps=distance_tolerance):
            return "MCLC3"
        elif cwt(expression, ">", 1, eps=distance_tolerance):
            return "MCLC5"


def _TRI_variation(k_alpha: float, k_beta: float, k_gamma: float, angle_tolerance=1e-4):
    r"""
    Four variations of the TRI lattice.

    Conditions :math:`k_{\alpha} \ne 90^{\circ}` and :math:`k_{\beta} \ne 90^{\circ}` are checked.

    :math:`\text{TRI}_{1a} k_{\alpha} > 90^{\circ}, k_{\beta} > 90^{\circ}, k_{\gamma} > 90^{\circ}, k_{\gamma} = \min(k_{\alpha}, k_{\beta}, k_{\gamma})`

    :math:`\text{TRI}_{1b} k_{\alpha} < 90^{\circ}, k_{\beta} < 90^{\circ}, k_{\gamma} < 90^{\circ}, k_{\gamma} = \max(k_{\alpha}, k_{\beta}, k_{\gamma})`

    :math:`\text{TRI}_{2a} k_{\alpha} > 90^{\circ}, k_{\beta} > 90^{\circ}, k_{\gamma} = 90^{\circ}`

    :math:`\text{TRI}_{2b} k_{\alpha} < 90^{\circ}, k_{\beta} < 90^{\circ}, k_{\gamma} = 90^{\circ}`

    Parameters
    ----------
    k_alpha : float
        Angle between reciprocal vectors :math:`b_2` and :math:`b_3`. In degrees.
    k_beta : float
        Angle between reciprocal vectors :math:`b_1` and :math:`b_3`. In degrees.
    k_gamma : float
        Angle between reciprocal vectors :math:`b_1` and :math:`b_2`. In degrees.
    angle_tolerance : float, default :math:`10^{-4}`
        Tolerance parameter for comparing two angles, given in degrees.

    Returns
    -------
    variation : str
        Variation of the lattice.
        Either "TRI1a", "TRI1b", "TRI2a" or "TRI2b".

    Raises
    ------
    ValueError
        If :math:`k_{\alpha} == 90^{\circ}` or :math:`k_{\beta} == 90^{\circ}` with given
        tolerance ``angle_tolerance``.
    """

    if cwt(k_alpha, "==", 90, eps=angle_tolerance) or cwt(
        k_beta, "==", 90, eps=angle_tolerance
    ):
        raise ValueError(
            f'(convention="SC"): TRI variation. k_alpha == 90 or k_beta == 90 with {angle_tolerance} tolerance.'
        )

    if cwt(k_gamma, "==", 90, eps=angle_tolerance):
        if cwt(k_alpha, ">", 90, eps=angle_tolerance) and cwt(
            k_beta, ">", 90, eps=angle_tolerance
        ):
            return "TRI2a"
        elif cwt(k_alpha, "<", 90, eps=angle_tolerance) and cwt(
            k_beta, "<", 90, eps=angle_tolerance
        ):
            return "TRI2b"
    elif cwt(min(k_gamma, k_beta, k_alpha), ">", 90, eps=angle_tolerance):
        return "TRI1a"
    elif cwt(max(k_gamma, k_beta, k_alpha), "<", 90, eps=angle_tolerance):
        return "TRI1b"
    else:
        return "TRI"


def sc_get_variation(cell, atoms, spglib_data=None):
    r"""
    Return variation of the lattice as defined in the paper by Setyawan and Curtarolo [1]_.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.
    atoms : dict
        Dictionary with N atoms. Expected keys:

        *   "positions" : (N, 3) |array-like|_

            Positions of the atoms in the basis of lattice vectors (``cell``). In other
            words - relative coordinates of atoms.
        *   "names" : (N, ) list of str, optional

            See Notes
        *   "species" : (N, ) list of str, optional

            See Notes
        *   "spglib_types" : (N, ) list of int, optional

            See Notes

        .. hint::
            Pass ``atoms = dict(positions=[[0, 0, 0]], spglib_types=[1])`` if you would
            like to interpret the ``cell`` alone (effectively assuming that the ``cell``
            is a primitive one).

    spglib_data : :py:class:`.SyntacticSugar`, optional
        If you need more control on the parameters passed to the spglib, then
        you can get ``spglib_data`` manually and pass it to this function.
        Use wulfric's interface to |spglib|_ as

        .. code-block:: python

            spglib_data = wulfric.get_spglib_data(...)

        using the same ``cell`` and ``atoms["positions"]`` that you are passing to this
        function.

    Returns
    -------
    variation : str
        Variation of the lattice defined by the ``cell``.

    Notes
    =====
    |spglib|_ uses ``types`` to distinguish the atoms. To see how wulfric deduces the
    ``types`` for given atoms see :py:func:`wulfric.get_spglib_types`.

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> # There is no variation of cubic lattice, therefore, just lattice type is
        >>> # returned
        >>> wulfric.crystal.sc_get_variation(
        ...     cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        ...     atoms=dict(positions=[[0, 0, 0]], spglib_types=[1]),
        ... )
        'CUB'
        >>> cell = wulfric.cell.sc_get_example("bct")
        >>> wulfric.crystal.sc_get_variation(
        ...     cell=cell,
        ...     atoms=dict(positions=[[0, 0, 0]], spglib_types=[1]),
        ... )
        'BCT1'

    """

    # Validate that the atoms dictionary is what expected of it
    validate_atoms(atoms=atoms, required_keys=["positions"], raise_errors=True)

    # Call spglib
    if spglib_data is None:
        spglib_data = get_spglib_data(cell=cell, atoms=atoms)
    # Or check that spglib_data were *most likely* produced via wulfric's interface
    elif not isinstance(spglib_data, SyntacticSugar):
        raise TypeError(
            f"Are you sure that spglib_data were produced via wulfric's interface? Expected SyntacticSugar, got {type(spglib_data)}."
        )
    # Validate that user-provided spglib_data match user-provided structure
    else:
        validate_spglib_data(cell=cell, atoms=atoms, spglib_data=spglib_data)

    cell = np.array(cell, dtype=float)

    lattice_type = spglib_data.crystal_family + spglib_data.centring_type

    angle_tolerance = (
        spglib_data.angle_tolerance if spglib_data.angle_tolerance != -1 else 1e-4
    )

    if lattice_type in ["tI", "oF", "hR", "mC", "aP"]:
        conv_cell, _ = get_conventional(
            cell=cell, atoms=atoms, convention="SC", spglib_data=spglib_data
        )
        conv_a, conv_b, conv_c, conv_alpha, _, _ = get_params(cell=conv_cell)

    if lattice_type == "tI":
        return _BCT_variation(conv_a, conv_c)

    if lattice_type == "oF":
        return _ORCF_variation(
            conv_a, conv_b, conv_c, distance_tolerance=spglib_data.symprec
        )

    if lattice_type == "hR":
        return _RHL_variation(conv_alpha, angle_tolerance=angle_tolerance)

    if lattice_type == "mC":
        prim_cell, _ = get_primitive(
            cell=cell, atoms=atoms, convention="SC", spglib_data=spglib_data
        )
        _, _, _, _, _, prim_k_gamma = get_params(get_reciprocal(cell=prim_cell))
        return _MCLC_variation(
            conv_a,
            conv_b,
            conv_c,
            conv_alpha,
            prim_k_gamma,
            distance_tolerance=spglib_data.symprec,
            angle_tolerance=angle_tolerance,
        )

    if lattice_type == "aP":
        _, _, _, k_alpha, k_beta, k_gamma = get_params(get_reciprocal(cell=conv_cell))
        return _TRI_variation(
            k_alpha,
            k_beta,
            k_gamma,
            angle_tolerance=angle_tolerance,
        )

    return SC_BRAVAIS_LATTICE_SHORT_NAMES[lattice_type]


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
import spglib
from copy import deepcopy
import numpy as np
from wulfric._syntactic_sugar import SyntacticSugar, add_sugar
from wulfric.crystal._crystal_validation import validate_atoms
from wulfric._exceptions import _raise_with_message, _SUPPORT_FOOTER
from wulfric.crystal._atoms import get_atom_species

from wulfric.constants._space_groups import CRYSTAL_FAMILY, CENTRING_TYPE

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def validate_spglib_data(cell, atoms, spglib_data) -> None:
    r"""
    Validate that ``cell`` and ``atoms["positions"]`` match the ones on which
    ``spglib_data`` was created.

    In details, it check that

    * ``cell`` is the same as ``spglib_data.original_cell``
    * ``atoms["positions"]`` are the same as ``spglib_data.original_positions``
    * ``wulfric.get_spglib_types(atoms=atoms)`` is the same as
      ``spglib_data.original_types``.

    Parameters
    ==========
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors. In the language of |spglib|_
        the same concept is usually called "basis vectors" or "lattice".
    atoms : dict
        Dictionary with N atoms. Expected keys:

        *   "positions" : (N, 3) |array-like|_

            Positions of the atoms in the basis of lattice vectors (``cell``). In other
            words - relative coordinates of atoms.
        *   "names" : (N, ) list of str, optional
        *   "species" : (N, ) list of str, optional
        *   "spglib_types" : (N, ) list of int, optional
    spglib_data : dict
        A dictionary with the added syntactic sugar (i.e. with the dot access to the keys),
        that is produced via call to :py:func:`.get_spglib_data`.

    Raises
    ======
    ValueError
        If ``cell`` and ``atoms`` do not match ``spglib_data``.
    """

    if not np.allclose(cell, spglib_data.original_cell):
        raise ValueError(
            "Validation of spglib_data against cell and atoms: cell mismatch."
        )

    if not np.allclose(atoms["positions"], spglib_data.original_positions):
        raise ValueError(
            "Validation of spglib_data against cell and atoms: atom's positions mismatch."
        )

    if get_spglib_types(atoms=atoms) != spglib_data.original_types:
        raise ValueError(
            "Validation of spglib_data against cell and atoms: atom's types mismatch."
        )


def get_spglib_types(atoms):
    r"""
    Constructs spglib_types for the given atoms.

    First satisfied rule is applied

    1.  "spglib_types" in atoms

        Return ``atoms["spglib_types"]``.

    2.  "species" in atoms.

        ``spglib_types`` are deduced from ``atoms["species"]``. If two atoms have the same
        species, then they will have the same integer assigned to them in
        ``spglib_types``.

    3.  "names" in ``atoms``

        Species are automatically deduced based on atom's names (via
        :py:func:`wulfric.crystal.get_atom_species`), and then the second rule is
        applied.

    Parameters
    ==========
    atoms : dict
        Dictionary with N atoms. At least one of the following keys is expected

        *   "names" : (N, ) list of str, optional
        *   "species" : (N, ) list of str, optional
        *   "spglib_types" : (N, ) list of int, optional

    Returns
    =======
    spglib_types : (N, ) list of int
        List of integer indices ready to be passed to |spglib|_.
    """

    validate_atoms(atoms=atoms, raise_errors=True)

    if "spglib_types" in atoms:
        spglib_types = atoms["spglib_types"]
    else:
        if "species" not in atoms and "names" in atoms:
            species = [
                get_atom_species(name=name, raise_on_fail=False)
                for name in atoms["names"]
            ]
        elif "species" in atoms:
            species = atoms["species"]
        else:
            raise ValueError(
                'Expected at least one of "spglib_types", "species" or "names" keys in ""atoms, found none.'
            )

        mapping = {
            name: index + 1 for index, name in enumerate(sorted(list(set(species))))
        }
        spglib_types = [mapping[name] for name in species]

    return spglib_types


def get_spglib_data(
    cell,
    atoms,
    spglib_symprec=1e-5,
    spglib_angle_tolerance=-1,
):
    r"""
    Interface to |spglib|_.

    The idea is that this is the only way to access the data from |spglib|_. In that way
    one can associate a dataset with a given ``cell`` and ``atoms`` and re-use it when necessary.

    Parameters
    ==========
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors. In the language of |spglib|_
        the same concept is usually called "basis vectors" or "lattice".
    atoms : dict
        Dictionary with N atoms. Expected keys:

        *   "positions" : (N, 3) |array-like|_
            Positions of the atoms in the basis of lattice vectors (``cell``). In other
            words - relative coordinates of atoms.
        *   "names" : (N, ) list of str, optional
            See Notes
        *   "species" : (N, ) list of str, optional
            See Notes
        *   "spglib_types" (N, ) list of int, optional
            See Notes

        .. hint::
            Pass ``atoms = dict(positions=[[0, 0, 0]], spglib_types=[1])`` if you would
            like to interpret the ``cell`` alone (effectively assuming that the ``cell``
            is a primitive one).
    spglib_symprec : float, default :math:`10^{-5}`
        Directly passed to |spglib|_. Tolerance parameter for the symmetry search.
    spglib_angle_tolerance : float, default -1
        Directly passed to |spglib|_. Tolerance parameter for the symmetry search.

    Returns
    =======
    spglib_data : dict
        A dictionary with the added syntactic sugar (i.e. with the dot access to the keys).

        Data that are included:

        * ``spglib_data.original_cell``

          Same as the given ``cell``

        * ``spglib_data.original_positions``

          Same as the given ``atoms["positions"]``

        * ``spglib_data.original_types``

          Same as ``wulfric.get_spglib_types(atoms=atoms)`` for given ``atoms``.

        * ``spglib_data.space_group_number``

          Number of the space group. ``1 <= spglib_data.space_group_number <= 230``.

        * ``spglib_data.crystal_family``

          Crystal family.

          * "c" for cubic
          * "h" for hexagonal
          * "t" for tetragonal
          * "o" for orhorhombic
          * "m" for monoclinic
          * "a" for triclinic

        * ``spglib_data.centring_type``

          Centring type.

          * "P" for primitive
          * "A" for side centered
          * "C" for side centered
          * "I" for body-centered
          * "R" for rhombohedral centring
          * "F" for all faces centered

        * ``spglib_data.conventional_cell``

          Conventional cell associated with the given structure in the same spatial
          orientation. In other words, it is a choice of the cell for the same crystal.
          It can contain more than one lattice point. Same as ``std_lattice`` of
          |spglib-dataset|_ but rotated back with the ``std_rotation_matrix`` of
          |spglib-dataset|_.

        * ``spglib_data.conventional_positions``

          N relative positions of the atoms in the basis of
          ``spglib_data.conventional_cell``. Same as ``std_positions`` of
          |spglib-dataset|_.

        * ``spglib_data.conventional_types``

          N types of the atoms. Same as ``std_types`` of |spglib-dataset|_.

        * ``spglib_data.primitive_cell``

          Primitive cell associated with the given structure in the same spatial
          orientation. In other words, it is a choice of the cell for the same crystal.
          It contains exactly one lattice point. Same as ``primitive_lattice``
          returned by |spglib-find-primitive|_, but rotated back with the
          ``std_rotation_matrix`` of |spglib-dataset|_.

        * ``spglib_data.primitive_positions``

          M relative positions of the atoms in the basis of
          ``spglib_data.primitive_cell``. Same as ``primitive_positions``
          returned by |spglib-find-primitive|_.

        * ``spglib_data.primitive_types``

          M types of the atoms. Same as ``primitive_types``
          returned by |spglib-find-primitive|_.

        * ``spglib_data.symprec`` angle_tolerance

          Tolerance parameter that was used to call |spglib|_.

        * ``spglib_data.angle_tolerance``

          Tolerance parameter that was used to call |spglib|_.

    Raises
    ======
    ValueError
        If some input data are not what is expected.
    TypeError
        If some input data are not what is expected.
    RuntimeError
        If spglib fail to detect symmetry.

    Notes
    =====
    |spglib|_ uses ``types`` to distinguish the atoms. To see how wulfric deduces
    ``types`` from given ``atoms`` see :py:func:`wulfric.get_spglib_types`.
    """

    try:
        # Validate input data
        # # Validate that the atoms dictionary is what expected of it
        validate_atoms(atoms=atoms, required_keys=["positions"], raise_errors=True)

        # Validate cell TODO: write _cell_validation.py,
        # perhaps check if it can form a parallelepiped
        try:
            cell = np.array(cell, dtype=float)
        except Exception as e:
            _raise_with_message(e=e, message=f"cell is not array-like, got\n{cell}")

        if cell.shape != (3, 3):
            raise ValueError(f"Expected shape of (3, 3) for cell, got {cell.shape}.")

        # Just a dictionary with dot-like access to its keys
        spglib_data = SyntacticSugar()

        # Populate with the input data
        spglib_data.original_cell = deepcopy(cell)
        spglib_data.original_positions = deepcopy(atoms["positions"])
        spglib_data.original_types = deepcopy(get_spglib_types(atoms=atoms))
        spglib_data.symprec = spglib_symprec
        spglib_data.angle_tolerance = spglib_angle_tolerance

        dataset = spglib.get_symmetry_dataset(
            (cell, spglib_data.original_positions, spglib_data.original_types),
            symprec=spglib_symprec,
            angle_tolerance=spglib_angle_tolerance,
        )

        if dataset is None:
            raise RuntimeError(
                f"spglib failed to detect symmetry for the given structure with spglib_symprec = {spglib_symprec} and spglib_angle_tolerance = {spglib_angle_tolerance}."
            )

        # For spglib <= 2.4.0
        if isinstance(dataset, dict):
            dataset = add_sugar(dataset)

        spglib_data.space_group_number = dataset.number
        spglib_data.crystal_family = CRYSTAL_FAMILY[dataset.number]
        spglib_data.centring_type = CENTRING_TYPE[dataset.number]
        # Rotate conventional cell back to the orientation of the given cell and atoms
        spglib_data.conventional_cell = (
            dataset.std_lattice @ dataset.std_rotation_matrix
        )
        spglib_data.conventional_positions = dataset.std_positions
        spglib_data.conventional_types = dataset.std_types

        primitive_cell, primitive_positions, primitive_types = spglib.find_primitive(
            (cell, spglib_data.original_positions, spglib_data.original_types),
            symprec=spglib_symprec,
            angle_tolerance=spglib_angle_tolerance,
        )

        # Rotate primitive cell back to the orientation of the given cell and atoms
        spglib_data.primitive_cell = primitive_cell @ dataset.std_rotation_matrix
        spglib_data.primitive_positions = primitive_positions
        spglib_data.primitive_types = primitive_types

        return spglib_data

    except Exception as e:
        _raise_with_message(
            e=e,
            message=f"Call to spglib failed. Spglib version {spglib.__version__}."
            + _SUPPORT_FOOTER,
        )


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
from copy import deepcopy
from typing import Iterable

import numpy as np

from wulfric.cell._basic_manipulation import get_reciprocal
from wulfric.constants._kpoints import HS_PLOT_NAMES
from wulfric.kpoints._path_and_points import (
    get_path_as_string,
    get_path_as_list,
    get_path_and_points,
)

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


class Kpoints:
    r"""
    Interface for convenient manipulations with the high-symmetry k-points and k-path in
    reciprocal space.

    Parameters
    ----------
    rcell : (3, 3) |array-like|_
        Reciprocal cell. Rows are interpreted as vectors.
    coordinates : list, optional
        Coordinates of high-symmetry points given in relative coordinates in reciprocal
        space (in the basis of ``rcell``).
    names: list, optional
        Names of the high-symmetry points. Used in ``path``. Has to have the same length
        as ``coordinates``. If ``None``, then use "K1", ... "KN", where
        ``N = len(coordinates)``.
    labels : list, optional
        List of the high-symmetry point's labels. Used for plotting. Has to have the same
        length as ``coordinates``. If ``None`` and ``names is None, then use "K$_1$", ...
        "K$_N$", where ``N = len(coordinates)``. If ``None`` but ``names are given, then
        ``labels = names``.
    path : str, optional
        K-path. Use elements of ``names`` to specify the path. If no names given, then use
        "K1-K2-...-KN", where ``N = len(coordinates)``.
    n : int
        Number of intermediate points between each pair of the high-symmetry points (high
        symmetry points excluded).

    Attributes
    ----------
    rcell : (3, 3) :numpy:`ndarray`
        Reciprocal cell. Rows are interpreted as vectors.
    hs_names : list
        Names of the high-symmetry points. Used in k-path and as main identifier of the
        point.
    hs_coordinates : dict
        Dictionary of the high-symmetry points coordinates. Coordinates are relative (in
        the basis of ``rcell``)

        .. code-block:: python

            {"name": [k_a, k_b, k_c], ... }

    hs_labels : dict
        Dictionary of labels for plotting.

        .. code-block:: python

            {"name": "label", ... }
    """

    def __init__(
        self, rcell, coordinates=None, names=None, labels=None, path=None, n=100
    ) -> None:
        self.rcell = np.array(rcell)

        if coordinates is None:
            coordinates = []

        if labels is None:
            if names is None:
                labels = [f"K$_{i + 1}$" for i in range(len(coordinates))]
            else:
                labels = [name for name in names]
        elif len(labels) != len(coordinates):
            raise ValueError(
                f"Amount of labels ({len(labels)}) does not match amount of points ({len(coordinates)})."
            )

        if names is None:
            names = [f"K{i + 1}" for i in range(len(coordinates))]
        elif len(names) != len(coordinates):
            raise ValueError(
                f"Amount of names ({len(names)}) does not match amount of points ({len(coordinates)})."
            )

        self.hs_coordinates = dict(
            [(names[i], np.array(coordinates[i])) for i in range(len(coordinates))]
        )
        self.hs_labels = dict([(names[i], labels[i]) for i in range(len(coordinates))])
        self.hs_names = names

        self._n = n

        self._path = None
        if path is None:
            path = "-".join(self.hs_names)
        self.path = path

    @staticmethod
    def from_crystal(
        cell,
        atoms,
        convention="HPKOT",
        with_time_reversal=True,
        n=100,
        spglib_data=None,
    ):
        r"""

        Parameters
        ----------
        cell : (3, 3) |array-like|_
            Matrix of a cell, rows are interpreted as vectors.
        atoms : dict
            Dictionary with N atoms. Expected keys:

            *   "positions" : (N, 3) |array-like|_

                Positions of the atoms in the basis of lattice vectors (``cell``). In other
                words - relative coordinates of atoms.
            *   "names" : (N, ) list of str, optional

                See Notes
            *   "species" : (N, ) list of str, optional

                See Notes
            *   "spglib_types" : (N, ) list of int, optional

                See Notes

            .. hint::
                Pass ``atoms = dict(positions=[[0, 0, 0]], spglib_types=[1])`` if you would
                like to interpret the ``cell`` alone (effectively assuming that the ``cell``
                is a primitive one).

        convention : str, default "HPKOT"
            Convention for the definition of the conventional cell. Case-insensitive.
            Supported:

            * "HPKOT" for [1]_
            * "SC" for [2]_
            * "spglib" for |spglib|_

        with_time_reversal : bool, default True
            Whether to assume that the system has time reversal symmetry. By default
            assumes that the system has it. The strategy for extending the path when
            ``with_time_reversal=False`` for the crystals without inversion symmetry is
            described in [1]_. For the systems with inversion symmetry this parameter does
            nothing.

            .. versionadded:: 0.6.3

        n : int, default 100
            Number of intermediate points between each pair of the high-symmetry points
            (high-symmetry points excluded).
        spglib_data : :py:class:`.SyntacticSugar`, optional
            If you need more control on the parameters passed to the spglib, then
            you can get ``spglib_data`` manually and pass it to this function.
            Use wulfric's interface to |spglib|_ as

            .. code-block:: python

                spglib_data = wulfric.get_spglib_data(...)

            using the same ``cell`` and ``atoms["positions"]`` that you are passing to this
            function.


        Notes
        -----
        |spglib|_ uses ``types`` to distinguish the atoms. To see how wulfric deduces the
        ``types`` for given atoms see :py:func:`wulfric.get_spglib_types`.

        If two atoms ``i`` and ``j`` have the same spglib_type (i.e.
        ``atoms["spglib_types"][i] == atoms["spglib_types"][j]``), but they have different
        property that is stored in ``atoms[key]`` (i.e ``atoms[key][i] != atoms[key][j]``),
        then those two atoms are considered equal. In the returned ``conventional_atoms``
        the value of the ``conventional_atoms[key]`` are populated based on the *last* found
        atom in ``atoms`` with each for spglib_type. This rule do not apply to the "positions"
        key.


        References
        ----------
        .. [1] Hinuma, Y., Pizzi, G., Kumagai, Y., Oba, F. and Tanaka, I., 2017.
            Band structure diagram paths based on crystallography.
            Computational Materials Science, 128, pp.140-184.
        .. [2] Setyawan, W. and Curtarolo, S., 2010.
            High-throughput electronic band structure calculations: Challenges and tools.
            Computational materials science, 49(2), pp. 299-312.
        """

        path, points = get_path_and_points(
            cell=cell,
            atoms=atoms,
            spglib_data=spglib_data,
            convention=convention,
            with_time_reversal=with_time_reversal,
            relative=True,
        )

        names = list(points.keys())
        coordinates = list(points.values())
        labels = [HS_PLOT_NAMES[name] for name in names]

        return Kpoints(
            rcell=get_reciprocal(cell=cell),
            coordinates=coordinates,
            names=names,
            labels=labels,
            path=path,
            n=n,
        )

    ################################################################################
    #                            High symmetry points                              #
    ################################################################################
    def add_hs_point(self, name, coordinate, label=None, relative=True) -> None:
        r"""
        Adds high-symmetry point.

        Parameters
        ----------
        name : str
            Name of the high-symmetry point.
        coordinate : (3,) array-like
            Coordinate of the high-symmetry point.
        label : str, optional
            Label of the high-symmetry point, ready to be plotted. If ``None``, then
            ``label = name``.
        relative : bool, optional
            Whether to interpret coordinates as relative or absolute.
        """

        if label is None:
            label = name

        if name in self.hs_names:
            raise ValueError(f'Point "{name}" already defined.')

        if not relative:
            # Transform from Cartesian coordinates to relative coordinates.
            coordinate = coordinate @ np.linalg.inv(self.rcell)

        self.hs_names.append(name)
        self.hs_coordinates[name] = np.array(coordinate)
        self.hs_labels[name] = label

    def remove_hs_point(self, name) -> None:
        r"""
        Removes high-symmetry point.

        Parameters
        ----------
        name : str
            Name of the high-symmetry point.
        """

        if name in self.hs_names:
            self.hs_names.remove(name)
            del self.hs_coordinates[name]
            del self.hs_labels[name]

    ################################################################################
    #                                Path attributes                               #
    ################################################################################
    @property
    def path(self) -> list:
        r"""
        K points path.

        Returns
        -------
        path : list of list of str
            K points path. Each subpath is a list of the high-symmetry points.

            .. code-block:: python

                path = [[K1, ..., KN], ...]
        """

        return self._path

    @path.setter
    def path(self, new_path):
        if isinstance(new_path, str):
            new_path = get_path_as_list(path_as_string=new_path)
        elif isinstance(new_path, Iterable):
            tmp_path = new_path
            new_path = []
            for subpath in tmp_path:
                if not isinstance(subpath, Iterable) or len(subpath) < 2:
                    raise ValueError(
                        f'Expected at least two points in each subpath, got "{subpath}"'
                    )

                subpath = [str(name) for name in subpath]
                new_path.append(subpath)

        # Check if all points are defined.
        for subpath in new_path:
            for point in subpath:
                if point not in self.hs_names:
                    raise ValueError(
                        (
                            f"Point '{point}' is not defined. Defined points are:\n  "
                            + "\n  ".join(
                                [
                                    f"{name} : {self.hs_coordinates[name]}"
                                    for name in self.hs_names
                                ]
                            )
                        )
                    )
        self._path = new_path

    @property
    def path_string(self) -> str:
        r"""
        K points path as a string.

        Returns
        -------
        path : str
        """

        return get_path_as_string(path_as_list=self._path)

    @property
    def n(self) -> int:
        r"""
        Amount of points between each pair of the high-symmetry points
        (high-symmetry points excluded).

        Returns
        -------
        n : int
        """

        return self._n

    @n.setter
    def n(self, new_n):
        if not isinstance(new_n, int):
            raise ValueError(
                f'n has to be integer. Given: {new_n} of type "{type(new_n)}"'
            )
        self._n = new_n

    ################################################################################
    #                         Attributes for the axis ticks                        #
    ################################################################################
    @property
    def labels(self) -> list:
        r"""
        Labels of high-symmetry points, ready to be plotted.

        For example for point "GAMMA" it returns R"$\Gamma$".

        If there are two high-symmetry points following one another in the path,
        it returns "X|Y" where X and Y are the labels of the two high-symmetry points.

        Returns
        -------
        labels : (N, ) list of str
            Labels, ready to be plotted. Same length as :py:attr:`.ticks`.
        """

        labels = []
        for s_i, subpath in enumerate(self.path):
            if s_i != 0:
                labels[-1] += "|" + self.hs_labels[subpath[0]]
            else:
                labels.append(self.hs_labels[subpath[0]])
            for name in subpath[1:]:
                labels.append(self.hs_labels[name])

        return labels

    def ticks(self, relative=False):
        r"""
        Tick's positions of the high-symmetry points, ready to be plotted.

        Same coordinated as in :py:meth:`.flat_points`.

        Parameters
        ----------
        relative : bool, default False
            Whether to use relative coordinates instead of the absolute ones.

        Returns
        -------
        ticks : (N, ) :numpy:`ndarray`
            Tick's positions, ready to be plotted. Same length as :py:attr:`.labels`.
        """

        if relative:
            cell = np.eye(3)
        else:
            cell = self.rcell

        ticks = []
        for s_i, subpath in enumerate(self.path):
            if s_i == 0:
                ticks.append(0)
            for i, name in enumerate(subpath[1:]):
                ticks.append(
                    np.linalg.norm(
                        self.hs_coordinates[name] @ cell
                        - self.hs_coordinates[subpath[i]] @ cell
                    )
                    + ticks[-1]
                )

        return np.array(ticks)

    ################################################################################
    #                   Points of the path with intermediate ones                  #
    ################################################################################
    def points(self, relative=False):
        r"""
        Coordinates of all points with n points between each pair of the high
        symmetry points (high-symmetry points excluded).

        Parameters
        ----------
        relative : bool, default False
            Whether to use relative coordinates instead of the absolute ones.

        Returns
        -------
        points : (N, 3) :numpy:`ndarray`
            Coordinates of all points.
        """

        if relative:
            cell = np.eye(3)
        else:
            cell = self.rcell

        points = None
        for subpath in self.path:
            for i in range(len(subpath) - 1):
                name = subpath[i]
                next_name = subpath[i + 1]
                new_points = np.linspace(
                    self.hs_coordinates[name] @ cell,
                    self.hs_coordinates[next_name] @ cell,
                    self._n + 2,
                )
                if points is None:
                    points = new_points
                else:
                    points = np.concatenate((points, new_points))
        return points

    # It can not just call for points and flatten them,
    # because it has to treat "|" as a special case.
    def flat_points(self, relative=False):
        r"""
        Flatten coordinates of all points with n points between each pair of the high
        symmetry points (high-symmetry points excluded).

        Used to plot band structure, dispersion, etc.

        Parameters
        ----------
        relative : bool, default False
            Whether to use relative coordinates instead of the absolute ones.

        Returns
        -------
        flat_points : (N, 3) :numpy:`ndarray`
            Flatten coordinates of all points.
        """

        if relative:
            cell = np.eye(3)
        else:
            cell = self.rcell

        flat_points = None
        for s_i, subpath in enumerate(self.path):
            for i in range(len(subpath) - 1):
                name = subpath[i]
                next_name = subpath[i + 1]
                points = (
                    np.linspace(
                        self.hs_coordinates[name] @ cell,
                        self.hs_coordinates[next_name] @ cell,
                        self._n + 2,
                    )
                    - self.hs_coordinates[name] @ cell
                )
                delta = np.linalg.norm(points, axis=1)
                if s_i == 0 and i == 0:
                    flat_points = delta
                else:
                    delta += flat_points[-1]
                    flat_points = np.concatenate((flat_points, delta))
        return flat_points

    ################################################################################
    #                                     Copy                                     #
    ################################################################################

    def copy(self):
        r"""
        Creates a copy of the kpoints.

        Returns
        -------
        kpoints : :py:class:`.Kpoints`
            Copy of the kpoints.
        """

        return deepcopy(self)

    ################################################################################
    #                                Human readables                               #
    ################################################################################

    def hs_table(self, decimals=8) -> str:
        r"""
        Table of the high-symmetry points.

        Parameters
        ----------
        decimals : int, optional
            Number of decimal places to round the coordinates.

        Returns
        -------
        table : str
            String with N+1 lines, where N is the amount of high-symmetry points.
            Each line contains the name of the high-symmetry point and its relative and
            absolute coordinates in a reciprocal space, i.e.::

                K1  0.0 0.0 0.0   0.0 0.0 0.0

            First line is a header::

                Name  rel_b1 rel_b2 rel_b3  k_x k_y k_z
        """

        d = decimals
        nd = max([len(name) for name in self.hs_names])
        table = [
            (
                f"{'Name':<{nd}}  "
                + f"{'rel_b1':>{d + 3}} "
                + f"{'rel_b2':>{d + 3}} "
                + f"{'rel_b3':>{d + 3}}  "
                + f"{'k_x':>{d + 3}} "
                + f"{'k_y':>{d + 3}} "
                + f"{'k_z':>{d + 3}}"
            )
        ]
        for name in self.hs_names:
            relative = self.hs_coordinates[name]
            i = f"{relative[0]: {d + 3}.{d}f}"
            j = f"{relative[1]: {d + 3}.{d}f}"
            k = f"{relative[2]: {d + 3}.{d}f}"
            absolute = self.hs_coordinates[name] @ self.rcell
            k_x = f"{absolute[0]: {d + 3}.{d}f}"
            k_y = f"{absolute[1]: {d + 3}.{d}f}"
            k_z = f"{absolute[2]: {d + 3}.{d}f}"
            table.append(f"{name:<{nd}}  {i} {j} {k}  {k_x} {k_y} {k_z}")
        return "\n".join(table)


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================

import pytest
from wulfric._spglib_interface import get_spglib_types


@pytest.mark.parametrize(
    "atoms, expected_types",
    [
        # Already present
        (dict(spglib_types=[1, 2, 2, 2, 3]), [1, 2, 2, 2, 3]),
        # Deduced from species
        (dict(species=["Cr", "Cr", "Fe"]), [1, 1, 2]),
        # Deduced from names
        (dict(names=["Cr1", "Cr2", "Fe"]), [1, 1, 2]),
        # All present
        (
            dict(
                spglib_types=[1, 1, 2],
                names=["Cr1", "Fe", "Cr2"],
                species=["Cr", "Fe", "Cr"],
            ),
            [1, 1, 2],
        ),
        # Names and species
        (
            dict(
                names=["Cr1", "Fe", "Cr2"],
                species=["Cr", "Cr", "Fe"],
            ),
            [1, 1, 2],
        ),
    ],
)
def test_get_spglib_types(atoms, expected_types):
    spglib_types = get_spglib_types(atoms=atoms)

    assert spglib_types == expected_types



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================

import sys

old_dir = set(dir())
old_dir.add("old_dir")

_SUPPORT_FOOTER = "\nPlease contact developers of wulfric (see https://docs.wulfric.org/en/latest/support.html)."


def _raise_with_message(e, message):
    # Python < 3.11
    if sys.version_info[1] < 11:
        args = e.args
        if not args:
            arg0 = message
        else:
            arg0 = f"{args[0]}\n{message}"
        e.args = (arg0,) + args[1:]
        raise e
    # Python >= 3.11
    else:
        e.add_note(message)
        raise e


class ConventionNotSupported(Exception):
    """
    Raised when the convention for the cell/crystall is not one of the supported ones.
    """

    def __init__(self, convention: str, supported_conventions: list):
        self.message = (
            f'Convention "{convention}" is not supported. Supported conventions are\n  * '
            + "\n  * ".join(supported_conventions)
        )

    def __str__(self):
        return self.message


class FailedToDeduceAtomSpecies(Exception):
    r"""
    Raised when the automatic deduction of the atom species from its name fails.
    """

    def __init__(self, name: str):
        self.message = f"Tried to deduce name from '{name}'. Failed."

    def __str__(self):
        return self.message


class NiggliReductionFailed(Exception):
    r"""
    Raised when niggli reduction reaches ``max_iterations``.
    """

    def __init__(self, max_iterations: int):
        self.message = f"Niggli reduction algorithm reached maximum amount of iterations: {max_iterations}"

    def __str__(self):
        return self.message


class PotentialBugError(Exception):
    def __init__(self, error_summary):
        self.message = (
            error_summary
            + "\nIf you see this error, than there might be a bug in wulfric."
            + _SUPPORT_FOOTER
        )

    def __str__(self):
        return self.message


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
import numpy as np
from wulfric._exceptions import _raise_with_message
from wulfric.constants._atoms import ATOM_SPECIES

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def validate_atoms(atoms, required_keys=None, raise_errors=True):
    r"""
    Validate keys and values of the atoms dictionary.

    Checks that all values in the atom's dictionary have the same amount of elements.
    Checks that atoms have required keys with expected shape of their values.

    If one of the following keys is in ``atoms``, then an extra check is performed on its
    value

    *   "names"

        Checks that the value is a list of N ``str``.

    *   "species"

        Checks that the value is a list of N ``str``.

        Checks that each element is a valid species of an atom (i.e. Cr, H, Br, Fe, ...) or "X".
        Note that "CR" is not a valid species, but "Cr" is.

    *   "positions"

        Checks that the ``atoms["positions"]`` has the shape (N, 3)

    *   "spglib_types"

        Checks that the value is a list of N ``int`` and each element is ``>= 1``.

    For all other keys checks that the values are iterables of the N elements each.

    Parameters
    ==========
    atoms : dict
        Dictionary of atoms.
    required_keys : list of str, optional
        List of required keys.
    raise_errors : bool, default True
        If ``False``, then no errors are raised.


    Returns
    =======
    check_passed : bool
        ``True`` if all checks passed. ``False`` otherwise.

    Raises
    ======
    TypeError
        If ``atoms[key]`` is not iterable for any ``key`` and ``raise_errors=False``.
    ValueError
        If any check is not passed and ``raise_errors=False``.
    """

    if required_keys is None:
        required_keys = []

    # Check that for every key value has N elements
    lengths = []
    for key in atoms:
        try:
            lengths.append(len(atoms[key]))
        except TypeError as e:
            if raise_errors:
                _raise_with_message(
                    e,
                    f'Failed to count elements of atoms["{key}"]. Are you sure that it is an iterable?\n'
                    + f'  atoms["{key}"] -> {atoms[key]}',
                )
            else:
                return False

    # Allow for an empty atoms dictionary
    if len(set(lengths)) not in [0, 1]:
        if raise_errors:
            raise ValueError(
                "Inconsistent amount of atoms:\n  * "
                + "\n  * ".join(
                    [
                        f'len(atoms["{key}"]) -> {N}'
                        for key, N in list(zip(list(atoms), lengths))
                    ]
                )
            )
        else:
            return False

    # Check that all required keys are present in atoms.
    for key in required_keys:
        if key not in atoms:
            if raise_errors:
                raise ValueError(
                    f'Expected to have the key "{key}" in atoms. Did not find one. Keys found in atoms:\n  * '
                    + "\n  * ".join(list(atoms))
                )
            else:
                return False

    # At this moment it is guaranteed that len(atoms[key]) is the same for all keys

    # Check values of positions
    if "positions" in atoms:
        try:
            shape = np.array(atoms["positions"]).shape
        except ValueError as e:
            if raise_errors:
                _raise_with_message(
                    e,
                    'atoms["positions"] is not array-like\n'
                    + f'  atoms["positions"] -> {atoms["positions"]}',
                )
            else:
                return False

        if len(shape) != 2 or shape[1] != 3:
            if raise_errors:
                raise ValueError(
                    f'Expected atoms["positions"] to have the shape (N, 3), got {shape}.\n'
                    + f'  atoms["positions"] -> {atoms["positions"]}'
                )
            else:
                return False

    # Check values of names
    if "names" in atoms:
        for index, element in enumerate(atoms["names"]):
            if not isinstance(element, str):
                if raise_errors:
                    raise ValueError(
                        f'Element #{index} of atoms["names"] is not a string:\n  '
                        f'atoms["names"][{index}] -> {element}'
                    )
                else:
                    return False

    # Check values of species
    if "species" in atoms:
        for index, element in enumerate(atoms["species"]):
            if not isinstance(element, str):
                if raise_errors:
                    raise ValueError(
                        f'Element #{index} of atoms["species"] is not a string:\n  '
                        f'atoms["species"][{index}] -> {element}'
                    )
                else:
                    return False

            if element != "X" and element not in ATOM_SPECIES:
                if raise_errors:
                    raise ValueError(
                        f'Element #{index} of atoms["species"] is not a valid species nor "X":\n  '
                        f'atoms["species"][{index}] -> {element}'
                    )
                else:
                    return False

    # Check spglib_types
    if "spglib_types" in atoms:
        for index, element in enumerate(atoms["spglib_types"]):
            if not isinstance(element, int):
                if raise_errors:
                    raise ValueError(
                        f'Element #{index} of atoms["spglib_types"] is not an integer:\n  '
                        f'atoms["spglib_types"][{index}] -> {element}'
                    )
                else:
                    return False

            if element < 1:
                if raise_errors:
                    raise ValueError(
                        f'Element #{index} of atoms["spglib_types"] is less than 1:\n  '
                        f'atoms["spglib_types"][{index}] -> {element}'
                    )
                else:
                    return False

    return True


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================

import numpy as np
from wulfric.crystal._crystal_validation import validate_atoms
from wulfric._exceptions import ConventionNotSupported, PotentialBugError
from wulfric._spglib_interface import get_spglib_data, validate_spglib_data
from wulfric._syntactic_sugar import SyntacticSugar
from wulfric.constants._sc_convention import SC_CONVENTIONAL_TO_PRIMITIVE
from wulfric.constants._hpkot_convention import HPKOT_CONVENTIONAL_TO_PRIMITIVE
from wulfric.crystal._conventional import get_conventional

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _get_unique(
    prim_cell,
    non_unique_positions,
    non_unique_types,
    repetition_number,
    distance_tolerance=1e-5,
):
    r"""
    Remove equivalent atoms from the primitive cell if any.

    Parameters
    ==========
    prim_cell : (3, 3) :numpy:`ndarray`
        Matrix of the cell, Rows are interpreted as vectors. Lattice vectors of the
        primitive cell.
    non_unique_positions : (N, 3) : :numpy:`ndarray`
        Positions of the atoms of the conventional cell in the basis of ``prim_cell``.
    non_unique_types : (N, ) list of int
        Types of atoms used to distinguish between them.
    repetition_number : int
        Correct amount of repetitions of each atom type in the conventional cell.
    distance_tolerance : float, default :math:`10^{-5}`
        Tolerance parameter for comparing two linear variables.

    Returns
    =======
    prim_positions : (M, 3) : :numpy:`ndarray`
        Unique atoms of the primitive cell. ``M = N / repetition_number``.
    prim_types : (M, ) list of int
        Types of atoms in the primitive cell.
    """

    # Get closest integer
    repetition_number = int(round(repetition_number, 0))

    non_unique_types = np.array(non_unique_types, dtype=int)

    # Deal with finite precision
    # Temporary solution
    non_unique_positions = np.round(non_unique_positions, decimals=8)

    # Move all to 000
    non_unique_positions = non_unique_positions % 1

    # Done twice it fixes some problems of finite precision arithmetics
    non_unique_positions = non_unique_positions % 1

    # Compute pair-wise distances between atoms
    distances = np.linalg.norm(
        non_unique_positions[:, np.newaxis, :] - non_unique_positions[np.newaxis, :, :],
        axis=2,
    )

    # Check if two atoms are the same for each pair
    same_atoms = np.isclose(
        distances, np.zeros(distances.shape), atol=distance_tolerance
    )

    # Count amount of equivalent atoms
    n_equiv = np.sum(same_atoms, axis=1)

    # Check that each atom has correct amount of twins
    if not (n_equiv == repetition_number * np.ones(n_equiv.shape, dtype=int)).all():
        abs_pos = non_unique_positions @ prim_cell
        raise ValueError(
            f"Some atoms have wrong number of twins. Expected {repetition_number} twins for each, got\n  * "
            + "\n  * ".join(
                [
                    f"atom at {abs_pos[i][0]:.5f} {abs_pos[i][1]:.5f} {abs_pos[i][2]:.5f} "
                    + f"({non_unique_positions[i][0]:.5f} {non_unique_positions[i][1]:.5f} {non_unique_positions[i][2]:.5f} @ prim_cell) "
                    + f"with type {non_unique_types[i]} has {n_equiv} twins"
                    for i in range(len(non_unique_positions))
                ]
            )
        )

    # Find a set of unique atoms
    N = len(non_unique_positions)
    available_atoms = np.ones(N).astype(bool)
    indices = np.linspace(0, N - 1, N, dtype=int)
    unique_atoms = np.zeros(N).astype(bool)
    i = 0
    while True:
        # Do not compare with itself
        available_atoms[i] = False
        # Remove all the same atoms for the future comparisons
        for j in indices[available_atoms]:
            if same_atoms[i][j]:
                available_atoms[j] = False
        # Every index that entered in this cycle is the first encountered atom of the equivalent group.
        unique_atoms[i] = True

        # Move to the next atom, that is not accounted for yet
        if available_atoms.any():
            i = indices[available_atoms][0]
        # Or end the cycle
        else:
            break

    return non_unique_positions[unique_atoms], non_unique_types[unique_atoms]


def get_primitive(cell, atoms, convention="HPKOT", spglib_data=None):
    r"""
    Return primitive cell and atoms associated with the given ``cell`` and ``atoms``.

    Parameters
    ==========
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.
    atoms : dict
        Dictionary with N atoms. Expected keys:

        *   "positions" : (N, 3) |array-like|_

            Positions of the atoms in the basis of lattice vectors (``cell``). In other
            words - relative coordinates of atoms.
        *   "names" : (N, ) list of str, optional

            See Notes
        *   "species" : (N, ) list of str, optional

            See Notes
        *   "spglib_types" : (N, ) list of int, optional

            See Notes

        .. hint::
            Pass ``atoms = dict(positions=[[0, 0, 0]], spglib_types=[1])`` if you would
            like to interpret the ``cell`` alone (effectively assuming that the ``cell``
            is a primitive one).

    convention : str, default "HPKOT"
        Convention for the definition of the primitive cell. Case-insensitive.
        Supported:

        * "HPKOT" for [1]_
        * "SC" for [2]_
        * "spglib" for |spglib|_ [3]_

    spglib_data : :py:class:`.SyntacticSugar`, optional
        If you need more control on the parameters passed to the spglib, then
        you can get ``spglib_data`` manually and pass it to this function.
        Use wulfric's interface to |spglib|_ as

        .. code-block:: python

            spglib_data = wulfric.get_spglib_data(...)

        using the same ``cell`` and ``atoms["positions"]`` that you are passing to this
        function.

    Returns
    =======
    primitive_cell : (3, 3) :numpy:`ndarray`
        Conventional cell.
    primitive_atoms : dict
        Dictionary of atoms of the conventional cell. Has all the same keys as the
        original ``atoms``. The values of each key are updated in such a way that
        ``primitive_cell`` with ``primitive_atoms`` describe the same crystal (and
        in the same spatial orientation) as ``cell`` with ``atoms``. It has all keys as
        in ``atoms``. Additional key ``"spglib_types"`` is added if it was not present in
        ``atoms``.

    See Also
    ========
    :ref:`user-guide_conventions_which-cell`
    wulfric.crystal.get_conventional
    wulfric.get_spglib_data


    Notes
    =====
    |spglib|_ uses ``types`` to distinguish the atoms. To see how wulfric deduces the
    ``types`` for given atoms see :py:func:`wulfric.get_spglib_types`.

    If two atoms ``i`` and ``j`` have the same spglib_type (i.e.
    ``atoms["spglib_types"][i] == atoms["spglib_types"][j]``), but they have different
    property that is stored in ``atoms[key]`` (i.e ``atoms[key][i] != atoms[key][j]``),
    then those two atoms are considered equal. In the returned ``primitive_atoms``
    the value of the ``primitive_atoms[key]`` are populated base on the *last* found
    atom in ``atoms`` with each for spglib_type. This rule do not apply to the "positions"
    key.


    References
    ==========
    .. [1] Hinuma, Y., Pizzi, G., Kumagai, Y., Oba, F. and Tanaka, I., 2017.
           Band structure diagram paths based on crystallography.
           Computational Materials Science, 128, pp.140-184.
    .. [2] Setyawan, W. and Curtarolo, S., 2010.
           High-throughput electronic band structure calculations: Challenges and tools.
           Computational materials science, 49(2), pp. 299-312.
    .. [3] Togo, A., Shinohara, K. and Tanaka, I., 2024.
           Spglib: a software library for crystal symmetry search.
           Science and Technology of Advanced Materials: Methods, 4(1), p.2384822.
    """

    # Validate that the atoms dictionary is what expected of it
    validate_atoms(atoms=atoms, required_keys=["positions"], raise_errors=True)

    # Call spglib
    if spglib_data is None:
        spglib_data = get_spglib_data(cell=cell, atoms=atoms)
    # Or check that spglib_data were *most likely* produced via wulfric's interface
    elif not isinstance(spglib_data, SyntacticSugar):
        raise TypeError(
            f"Are you sure that spglib_data were produced via wulfric's interface? Expected SyntacticSugar, got {type(spglib_data)}."
        )
    # Validate that user-provided spglib_data match user-provided structure
    else:
        validate_spglib_data(cell=cell, atoms=atoms, spglib_data=spglib_data)

    convention = convention.lower()

    # Straightforward interface to spglib
    if convention == "spglib":
        prim_cell = spglib_data.primitive_cell
        prim_positions = spglib_data.primitive_positions
        prim_types = spglib_data.primitive_types
    # Some work needed for other conventions
    elif convention in ["hpkot", "sc"]:
        lattice_type = spglib_data.crystal_family + spglib_data.centring_type
        conv_cell, conv_atoms = get_conventional(
            cell=cell, atoms=atoms, convention=convention, spglib_data=spglib_data
        )

        if convention == "hpkot":
            matrix = HPKOT_CONVENTIONAL_TO_PRIMITIVE[lattice_type]
        elif convention == "sc":
            matrix = SC_CONVENTIONAL_TO_PRIMITIVE[lattice_type]
        else:
            raise PotentialBugError('get_primitive(convention="HPKOT/SC").')

        prim_cell = matrix.T @ conv_cell

        non_unique_positions = (
            conv_atoms["positions"] @ conv_cell @ np.linalg.inv(prim_cell)
        )

        # Conventional cell may contain more atoms than primitive one, thus one needs to
        # remove equivalent atoms
        prim_positions, prim_types = _get_unique(
            prim_cell=prim_cell,
            non_unique_positions=non_unique_positions,
            non_unique_types=conv_atoms["spglib_types"],
            repetition_number=abs(np.linalg.det(conv_cell) / np.linalg.det(prim_cell)),
            distance_tolerance=spglib_data.symprec,
        )
    else:
        raise ConventionNotSupported(
            convention, supported_conventions=["HPKOT", "SC", "spglib"]
        )

    # Create primitive atoms
    prim_atoms = dict(positions=prim_positions)

    # Get mapping from original atoms to primitive ones through types
    types_mapping = {
        type_index: index for index, type_index in enumerate(spglib_data.original_types)
    }

    # Populate primitive_atoms with all keys that have been defined in the original atoms.
    for key in atoms:
        if key != "positions":
            prim_atoms[key] = []
            for type_index in prim_types:
                prim_atoms[key].append(atoms[key][types_mapping[type_index]])

    # Add spglib_types to new atoms if necessary
    if "spglib_types" not in prim_atoms:
        prim_atoms["spglib_types"] = prim_types

    return prim_cell, prim_atoms


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
import os
from wulfric.io._vasp import load_poscar
from wulfric.constants._hpkot_convention import HPKOT_EXTENDED_BL_SYMBOLS

__all__ = ["hpkot_get_example"]


def hpkot_get_example(extended_bl_symbol, with_inversion=False):
    r"""
    Returns an example of the crystal structure for each of the
    given extended Bravais lattice symbol as defined in [1]_.

    .. versionadded:: 0.6.3

    Crystal structures are taken from the |seekpath|_ repository
    with the permission of its authors.

    See :ref:`api_constants_HPKOT_EXTENDED_BL_SYMBOLS` for the full list of supported
    symbols.

    .. warning::
        "oF2", "oI2", "oA1", "oA2" with ``with_inversion=True`` are not supported fo now.


    Parameters
    ----------
    extended_bl_symbol : str
        Extended Bravais lattice symbol. Case-sensitive.
    with_inversion : bool, default False
        Whether give an example that has inversion symmetry.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Unit cell of the crystal structure.
    atoms : dict
        Atoms of the crystal structure.

    Raises
    ------
    ValueError
        If ``extended_bl_symbol`` is not supported.


    References
    ----------
    .. [1] Hinuma, Y., Pizzi, G., Kumagai, Y., Oba, F. and Tanaka, I., 2017.
           Band structure diagram paths based on crystallography.
           Computational Materials Science, 128, pp.140-184.
    """

    if extended_bl_symbol in ["oF2", "oI2", "oA1", "oA2"] and with_inversion:
        raise ValueError(
            '"oF2", "oI2", "oA1", "oA2" with with_inversion=True currently are not supported.'
        )

    supported_symbols = list(HPKOT_EXTENDED_BL_SYMBOLS)

    supported_symbols += list(set([_[:2] for _ in supported_symbols]))

    if extended_bl_symbol not in supported_symbols:
        raise ValueError(
            f'Extended Bravais lattice symbol "{extended_bl_symbol}" is not supported. Supported are:\n  * '
            + "\n  * ".join(supported_symbols)
        )

    if len(extended_bl_symbol) == 2:
        if extended_bl_symbol == "aP":
            extended_bl_symbol = "aP2"
        else:
            extended_bl_symbol += "1"

    POSCAR_filename = os.path.join(
        os.path.split(os.path.abspath(__file__))[0], "POSCAR-files", extended_bl_symbol
    )

    if with_inversion:
        POSCAR_filename = os.path.join(POSCAR_filename, "POSCAR_inversion")
    else:
        POSCAR_filename = os.path.join(POSCAR_filename, "POSCAR_noinversion")

    cell, atoms, _ = load_poscar(file_object=POSCAR_filename)

    return cell, atoms
